# (c) Copyright 2013, 2014, University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
import datetime
import logging
import json
import hydra_base as hb
from hydra_base.lib.objects import JSONObject
from hydra_base.exceptions import ResourceNotFoundError

from hydra_base.util import flatten_dict, count_levels

log = logging.getLogger(__name__)


@pytest.fixture()
def relative_dataframe():
    """
        Create a timeseries which has relative timesteps:
        1, 2, 3 as opposed to timestamps
    """
    t1 = 1
    t2 = 2
    t3 = 3
    val_1 = [[[11, 22, 33], [55, 44, 66]], [[100, 200, 300], [400, 500, 600]], [[99, 88, 77], [66, 55, 44]]]
    val_2 = ["1.1", "2.2", "3.3"]
    val_3 = ["3.3", "", ""]

    timeseries = json.dumps({0: {t1: val_1, t2: val_2, t3: val_3}})

    return timeseries


@pytest.fixture()
def arbitrary_dataframe():
    """
        Create a timeseries which has relative timesteps:
        1, 2, 3 as opposed to timestamps
    """
    t1 = 'arb'
    t2 = 'it'
    t3 = 'rary'
    val_1 = [[[0.1, 0.2, 33], [0.5, 0.4, 0.6]], [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]], [[0.9,0.8,0.7],[0.6,0.5,0.4]]]
    val_2 = ["1.0", "2.0", "3.0"]
    val_3 = ["3.0", "", ""]

    timeseries = json.dumps({0:{t1:val_1, t2:val_2, t3:val_3}})

    return timeseries


@pytest.fixture()
def seasonal_timeseries():
    """
        Create a timeseries which has relative timesteps:
        1, 2, 3 as opposed to timestamps
    """
    t1 ='9999-01-01'
    t2 ='9999-02-01'
    t3 ='9999-03-01'
    val_1 = [[[1, 2, 33], [5, 4, 6]], [[10, 20, 30], [40, 50, 60]], [[9,8,7],[6,5,4]]]

    val_2 = ["1.0", "2.0", "3.0"]
    val_3 = ["3.0", "", ""]

    timeseries = json.dumps({0: {t1: val_1, t2: val_2, t3: val_3}})

    return timeseries


class TestTimeSeries:
    """
        Test for timeseries-based functionality
    """
    #@pytest.mark.xfail(reason='Relative timesteps are being converted to timestamps. ')
    def test_relative_dataframe(self, client, network, relative_dataframe):
        """
            Test for relative timeseries for example, x number of hours from hour 0.
        """

        s = network['scenarios'][0]
        assert len(s['resourcescenarios']) > 0
        for rs in s['resourcescenarios']:
            if rs['dataset']['type'] == 'dataframe':
                rs['dataset']['value'] = relative_dataframe

        new_network_summary = client.add_network(network)
        new_net = client.get_network(new_network_summary.id, include_data=True)

        new_s = new_net['scenarios'][0]
        new_rss = new_s['resourcescenarios']

        assert len(new_rss) > 0
        for new_rs in new_rss:
            if new_rs.dataset.type == 'dataframe':
                ret_ts_dict = list(json.loads(new_rs.dataset.value).values())[0]
                client_ts   = json.loads(relative_dataframe)['0']
                for new_timestep in client_ts.keys():
                    # TODO this bit appears broken. Somewhere in hydra the timesteps
                    # are convert to timestamps.
                    assert ret_ts_dict.get(new_timestep) == client_ts[new_timestep]


    def test_arbitrary_dataframe(self, client, network, arbitrary_dataframe):

        s = network['scenarios'][0]
        for rs in s['resourcescenarios']:
            if rs['dataset']['type'] == 'dataframe':
                rs['dataset']['value'] = arbitrary_dataframe

        new_network_summary = client.add_network(network)
        new_net = client.get_network(new_network_summary.id, include_data=True)

        new_s = new_net.scenarios[0]
        new_rss = new_s.resourcescenarios
        for new_rs in new_rss:
            if new_rs.dataset.type == 'dataframe':
                ret_ts_dict = {}
                ret_ts_dict = list(json.loads(new_rs.dataset.value).values())[0]
                client_ts   = json.loads(arbitrary_dataframe)['0']
                for new_timestep in client_ts.keys():
                    assert ret_ts_dict[new_timestep] == client_ts[new_timestep]

    def test_get_relative_data_between_times(self, client, network, relative_dataframe):

        # TODO The following is shared with `test_relative_dataframe` it could be put in a fixture
        s = network['scenarios'][0]
        assert len(s['resourcescenarios']) > 0
        for rs in s['resourcescenarios']:
            if rs['dataset']['type'] == 'dataframe':
                rs['dataset']['value'] = relative_dataframe

        new_network_summary = client.add_network(network)
        new_net = client.get_network(new_network_summary.id, include_data=True)
        scenario = new_net.scenarios[0]
        val_to_query = None
        for rs in scenario.resourcescenarios:
            if rs.dataset.type == 'dataframe':
                val_to_query = rs.dataset
                break

        now = datetime.datetime.now()
        x = client.get_vals_between_times(
            val_to_query.id,
            0,
            5,
            None,
            0.5,
            )
        assert len(x.data) > 0

        invalid_qry = client.get_vals_between_times(
            val_to_query.id,
            now,
            now + datetime.timedelta(minutes=75),
            'minutes',
            5
            )

        assert eval(invalid_qry.data) == []

    def test_seasonal_timeseries(self, client, network, seasonal_timeseries):

        s = network['scenarios'][0]
        for rs in s['resourcescenarios']:
            if rs['dataset']['type'] == 'timeseries':
                rs['dataset']['value'] = seasonal_timeseries

        new_network_summary = client.add_network(network)
        new_net = client.get_network(new_network_summary.id, include_data=True)

        scenario = new_net.scenarios[0]
        val_to_query = None
        for rs in scenario.resourcescenarios:
            if rs.dataset.type == 'timeseries':
                val_to_query = rs.dataset
                break

        val_a = list(json.loads(val_to_query.value).values())[0]

        jan_val = client.get_val_at_time(
            val_to_query.id,
            [datetime.datetime(2000, 1, 10, 00, 00, 00)]
           )
        feb_val = client.get_val_at_time(
            val_to_query.id,
            [datetime.datetime(2000, 2, 10, 00, 00, 00)]
           )
        mar_val = client.get_val_at_time(
            val_to_query.id,
            [datetime.datetime(2000, 3, 10, 00, 00, 00)]
           )
        oct_val = client.get_val_at_time(
            val_to_query.id,
            [datetime.datetime(2000, 10, 10, 00, 00, 00)]
           )

        local_val = list(json.loads(val_to_query.value).values())[0]
        assert json.loads(jan_val.data) == local_val['9999-01-01']
        assert json.loads(feb_val.data) == local_val['9999-02-01']
        assert json.loads(mar_val.data) == local_val['9999-03-01']
        assert json.loads(oct_val.data) == local_val['9999-03-01']

        start_time = datetime.datetime(2000, 7, 10, 00, 00, 00)
        vals = client.get_vals_between_times(
            val_to_query.id,
            start_time,
            start_time + datetime.timedelta(minutes=75),
            'minutes',
            1,
            )

        data = json.loads(vals.data)
        original_val = local_val['9999-03-01']
        assert len(data) == 76
        for val in data:
            assert original_val == val

    def test_multiple_vals_at_time(self, client, network_with_data, seasonal_timeseries):

        s = network_with_data['scenarios'][0]

        for rs in s['resourcescenarios']:
            if rs['dataset']['type'] == 'timeseries':
                rs['dataset']['value'] = seasonal_timeseries

        client.update_network(network_with_data)
        new_net = client.get_network(network_with_data.id, include_data=True)

        scenario = new_net.scenarios[0]
        val_to_query = None
        for rs in scenario.resourcescenarios:
            if rs.dataset.type == 'timeseries':
                val_to_query = rs.dataset
                break

        val_a = json.loads(val_to_query.value)

        dtformat = hb.config.get('DEFAULT', 'datetime_format', "%Y-%m-%dT%H:%M:%S.%f000Z")
        fmt = datetime.datetime.strftime
        qry_times = [
            fmt(datetime.datetime(2000, 1, 10, 00, 00, 00), dtformat),
            fmt(datetime.datetime(2000, 2, 10, 00, 00, 00), dtformat),
            fmt(datetime.datetime(2000, 3, 10, 00, 00, 00), dtformat),
            fmt(datetime.datetime(2000, 10, 10, 00, 00, 00), dtformat),
            ]

        seasonal_vals = client.get_multiple_vals_at_time(
            [val_to_query.id],
            qry_times,
           )

        #TODO: Figure out why the mysqlclient library with python 2.7.14 was causing this to break
        return_val = seasonal_vals['dataset_%s'%int(val_to_query.id)]

        dataset_vals = val_a['0']

        assert return_val[qry_times[0]] == dataset_vals['9999-01-01']
        assert return_val[qry_times[1]] == dataset_vals['9999-02-01']
        assert return_val[qry_times[2]] == dataset_vals['9999-03-01']
        assert return_val[qry_times[3]] == dataset_vals['9999-03-01']

        start_time = datetime.datetime(2000, 7, 10, 00, 00, 00)
        vals = client.get_vals_between_times(
            val_to_query.id,
            start_time,
            start_time + datetime.timedelta(minutes=75),
            'minutes',
            1,
            )

        data = json.loads(vals.data)
        original_val = dataset_vals['9999-03-01']
        assert len(data) == 76
        for val in data:
            assert original_val == val

    def test_get_data_between_times(self, client, network_with_data):

        # Convenience renaming
        net = network_with_data

        scenario = net.scenarios[0]
        val_to_query = None
        for rs in scenario.resourcescenarios:
            if rs.dataset.type == 'timeseries':
                val_to_query = rs.dataset
                break

        val_a, val_b = list(json.loads(val_to_query.value)['test_column'].values())[:2]

        now = datetime.datetime.now(datetime.timezone.utc)

        vals = client.get_vals_between_times(
            val_to_query.id,
            now,
            now + datetime.timedelta(minutes=75),
            'minutes',
            1,
            )

        data = json.loads(vals.data)
        assert len(data) == 76
        for val in data[60:75]:
            x = val_b
            assert x == val_b
        for val in data[0:59]:
            x = val_a
            assert x == val_a

    def test_descriptor_get_data_between_times(self, client, network_with_data):
        net = network_with_data
        scenario = net.scenarios[0]
        val_to_query = None
        for rs in scenario.resourcescenarios:
            if rs.dataset.type == 'descriptor':
                val_to_query = rs.dataset
                break

        now = datetime.datetime.now()

        value = client.get_vals_between_times(
            val_to_query.id,
            now,
            now + datetime.timedelta(minutes=75),
            'minutes',
            1
            )

        assert json.loads(value.data) == ['test']



#Commented out because an imbalanced array is now allowed. We may add checks
#for this at a later date if needed, but for now we are going to leave such
#validation up to the client.
#class ArrayTest(server.HydraBaseTest):
#    def test_array_format(self):
#        bad_net = self.build_network()
#
#        s = bad_net['scenarios'][0]
#        for rs in s['resourcescenarios']:
#            if rs['value']['type'] == 'array':
#                rs['value']['value'] = json.dumps([[1, 2] ,[3, 4, 5]])
#
#        self.assertRaises(WebFault, client.add_network,bad_net)
#
#        net = self.build_network()
#        n = client.add_network(net)
#        good_net = client.get_network(n.id)
#
#        s = good_net.scenarios[0]
#        for rs in s.resourcescenarios:
#            if rs.value.type == 'array':
#                rs.value.value = json.dumps([[1, 2] ,[3, 4, 5]])
#                #Get one of the datasets, make it uneven and update it.
#                self.assertRaises(WebFault, client.update_dataset,rs)

@pytest.fixture
def collection_json_object():
    collection = JSONObject(dict(
        type="descriptor",
        name="Test collection"
    ))
    return collection


@pytest.fixture()
def network_with_dataset_collection(client, network_with_data, collection_json_object):
    network = network_with_data

    scenario_id = network.scenarios[0].id
    scenario_data = client.get_scenario_data(scenario_id)

    collection = collection_json_object

    group_dataset_ids = [scenario_data[0].id, ]
    for d in scenario_data:
        if d.type == 'timeseries' and d.id not in group_dataset_ids:
            group_dataset_ids.append(d.id)
            break

    collection.dataset_ids = group_dataset_ids
    collection.name = 'test soap collection %s' % (datetime.datetime.now())

    newly_added_collection = client.add_dataset_collection(collection)
    return newly_added_collection


class TestDataCollection:

    def test_get_collections_like_name(self, client, network_with_dataset_collection):
        collections = client.get_collections_like_name('test')
        assert len(collections) > 0, "collections were not retrieved correctly!"

    def test_get_collection_datasets(self, client, network_with_dataset_collection):
        collections = client.get_collections_like_name('test')
        datasets = client.get_collection_datasets(collections[-1]['id'])

        assert len(datasets) > 0, "Datasets were not retrieved correctly!"

    def test_add_collection(self, client, network_with_dataset_collection):
        collection = network_with_dataset_collection

        assert collection.id is not None, "Dataset collection does not have an ID!"
        assert len(collection['items']) == 2, "Dataset collection does not have any items!"

    def test_delete_collection(self, client, network_with_dataset_collection):

        collection = network_with_dataset_collection

        # Get all collections and make sure this collection is present
        all_collections_pre = client.get_all_dataset_collections()

        all_collection_ids_pre = [c.id for c in all_collections_pre]

        assert collection.id in all_collection_ids_pre

        # Delete the collection
        client.delete_dataset_collection(collection.id)

        # Get all the collections again and make sure the deleted collection is not present
        all_collections_post = client.get_all_dataset_collections()
        all_collection_ids_post = [c.id for c in all_collections_post]

        assert collection.id not in all_collection_ids_post

        with pytest.raises(hb.exceptions.HydraError):
            client.get_dataset_collection(collection.id)

    def test_get_all_collections(self, client, network_with_dataset_collection):

        collection = network_with_dataset_collection

        collections = client.get_all_dataset_collections()
        assert collection.id in [dc.id for dc in collections]

    def test_add_dataset_to_collection(self, client, network_with_data, collection_json_object):

        network = network_with_data

        scenario_id = network.scenarios[0].id

        scenario_data = client.get_scenario_data(scenario_id)

        collection = collection_json_object

        group_dataset_ids = [scenario_data[0].id, ]
        for d in scenario_data:
            if d.type == 'timeseries' and d.id not in group_dataset_ids:
                group_dataset_ids.append(d.id)
                break

        dataset_id_to_add = None
        for d in scenario_data:
            if d.type == 'array' and d.id not in group_dataset_ids:
                dataset_id_to_add = d.id
                break

        collection.dataset_ids = group_dataset_ids
        collection.name = 'test soap collection %s'%(datetime.datetime.now())

        newly_added_collection = client.add_dataset_collection(collection)

        previous_dataset_ids = []
        for item in newly_added_collection['items']:
            previous_dataset_ids.append(item.dataset_id)

        # This acts as a test for the 'check_dataset_in_collection' code
        assert client.check_dataset_in_collection(dataset_id_to_add, newly_added_collection.id) == 'N'
        assert client.check_dataset_in_collection(99999, newly_added_collection.id) == 'N'

        with pytest.raises(hb.exceptions.HydraError):
            client.check_dataset_in_collection(99999, 99999)

        client.add_dataset_to_collection(dataset_id_to_add, newly_added_collection.id)

        assert client.check_dataset_in_collection(dataset_id_to_add, newly_added_collection.id) == 'Y'

        updated_collection = client.get_dataset_collection(newly_added_collection.id)

        new_dataset_ids = []
        for item in updated_collection['items']:
            new_dataset_ids.append(item.dataset_id)

        assert set(new_dataset_ids) - set(previous_dataset_ids) == set([dataset_id_to_add])

    def test_add_datasets_to_collection(self, client, network_with_data, collection_json_object):

        network = network_with_data

        scenario_id = network.scenarios[0].id

        scenario_data = client.get_scenario_data(scenario_id)

        collection = collection_json_object

        group_dataset_ids = [scenario_data[0].id, ]
        for d in scenario_data:
            if d.type == 'timeseries' and d.id not in group_dataset_ids:
                group_dataset_ids.append(d.id)
                break

        dataset_ids_to_add = []
        for d in scenario_data:
            if d.type == 'array' and d.id not in group_dataset_ids:
                dataset_ids_to_add.append(d.id)

        collection.dataset_ids = group_dataset_ids
        collection.name = 'test soap collection %s'%(datetime.datetime.now())

        newly_added_collection = client.add_dataset_collection(collection)

        previous_dataset_ids = []
        for item in newly_added_collection['items']:
            previous_dataset_ids.append(item.dataset_id)

        client.add_datasets_to_collection(dataset_ids_to_add, newly_added_collection.id)

        updated_collection = client.get_dataset_collection(newly_added_collection.id)

        new_dataset_ids = []
        for item in updated_collection['items']:
            new_dataset_ids.append(item.dataset_id)

        assert set(new_dataset_ids) - set(previous_dataset_ids) == set(dataset_ids_to_add)

    def test_remove_dataset_from_collection(self, client, network_with_data, collection_json_object):

        network = network_with_data

        scenario_id = network.scenarios[0].id
        scenario_data = client.get_scenario_data(scenario_id)

        collection = collection_json_object
        dataset_id = scenario_data[0].id
        group_dataset_ids = [dataset_id, ]
        for d in scenario_data:
            if d.type == 'timeseries' and d.id not in group_dataset_ids:
                group_dataset_ids.append(d.id)
                break

        collection.dataset_ids = group_dataset_ids
        collection.name = 'test soap collection %s' % (datetime.datetime.now())

        collection = client.add_dataset_collection(collection)

        previous_dataset_ids = []
        for item in collection['items']:
            previous_dataset_ids.append(item.dataset_id)

        client.remove_dataset_from_collection(dataset_id, collection.id)

        updated_collection = client.get_dataset_collection(collection.id)

        new_dataset_ids = []
        for item in updated_collection['items']:
            new_dataset_ids.append(item.dataset_id)

        assert set(previous_dataset_ids) - set(new_dataset_ids) == set([dataset_id])

    def test_delete_dataset_thats_in_a_collection(self, client, network_with_data, collection_json_object):

        network = network_with_data

        scenario_id = network.scenarios[0].id
        scenario_data = client.get_scenario_data(scenario_id)

        collection = collection_json_object
        dataset_id = None
        group_dataset_ids = []
        for d in scenario_data:
            if dataset_id is None and d.type == 'timeseries':
                dataset_id = d.id
                group_dataset_ids.append(d.id)
            elif d.type == 'timeseries' and d.id not in group_dataset_ids:
                group_dataset_ids.append(d.id)

        collection.dataset_ids = group_dataset_ids
        collection.name = 'test soap collection %s' % (datetime.datetime.now())

        newly_added_collection = client.add_dataset_collection(collection)

        # Make dataset_id into an orphaned dataset.
        for s in network.scenarios:
            for rs in s.resourcescenarios:
                if rs.dataset_id == dataset_id:
                    client.delete_resourcedata(scenario_id, rs)

        new_collection = client.get_dataset_collection(newly_added_collection.id)

        new_dataset_ids = []

        for item in new_collection['items']:
            new_dataset_ids.append(item.dataset_id)

        assert dataset_id in new_dataset_ids

        client.delete_dataset(dataset_id)

        updated_collection = client.get_dataset_collection(newly_added_collection.id)
        updated_dataset_ids = []
        for item in updated_collection['items']:
            updated_dataset_ids.append(item.dataset_id)

        assert dataset_id not in updated_dataset_ids

class TestSharing:

    def test_hide_data(self, client, network_with_data):
        """
            Test for the hiding of data.
            Create a network with some data.
            Hide the timeseries created, check if another user can see it.
            Share the time series with one users. Check if they can see it but a third user can't.
        """

        #Let User B view network 1, but not edit it (read_only is 'Y')
        client.share_network(network_with_data.id, ["UserB", "UserC"], 'Y', 'N')

        data_to_hide = network_with_data.scenarios[0].resourcescenarios[-1].dataset.id

        client.hide_dataset(data_to_hide, ["UserB"], 'Y', 'Y', 'Y')

        client.user_id = pytest.user_c.id
        netA = client.get_network(network_with_data.id, include_data=True)

        for rs in netA.scenarios[0].resourcescenarios:
            d = rs.dataset
            if d.id == data_to_hide:
                assert d.hidden == 'Y'
                assert d.value is None
            else:
                #The rest of the data is unhidden, so should be there.
                assert d.hidden == 'N'
                assert d.value is not None

        netB = client.get_network(network_with_data.id, include_data=True)

        for rs in netB.scenarios[0].resourcescenarios:
            d = rs.dataset
            if d.id == data_to_hide:
                assert d.hidden == 'Y'
                assert d.value is None
            else:
                #The rest of the data is unhidden, so should be there.
                assert d.hidden == 'N'
                assert d.value is not None

        directly_retrieved_scenario = client.get_scenario(netB.scenarios[0].id)

        for rs in directly_retrieved_scenario.resourcescenarios:
            d = rs.dataset
            if d.id == data_to_hide:
                assert d.hidden == 'Y'
                assert d.value is None
            else:
                #The rest of the data is unhidden, so should be there.
                assert d.hidden == 'N'
                assert d.value is not None

        scenario_data = client.get_scenario_data(netB.scenarios[0].id)

        for d in scenario_data:
            if d.id == data_to_hide:
                assert d.hidden == 'Y'
                assert d.value is None
            else:
                #The rest of the data is unhidden, so should be there.
                assert d.hidden == 'N'
                assert d.value is not None

        client.user_id = pytest.root_user_id

    def test_replace_hidden_data(self, client, network_with_data):
        """
            test_replace_hidden_data
            Test for the case where one user hides data and another
            user sets the data to something else.

            User A Creates a network with some data
            User A Hides the timeseries created.
            User A shares network with User B

            Check user B cannot see timeseries value
            User B creates a new timeseries, and replaces the hidden one.
            Save network.
            Attribute now should have a new, unhidden dataset assigned to that attribute.
        """

        #Let User B view network 1, but not edit it (read_only is 'Y')
        client.share_network(network_with_data.id, ["UserB", "UserC"], 'N', 'N')

        for d in  network_with_data.scenarios[0].resourcescenarios:
            if d.dataset.type == 'timeseries':
                attr_to_be_changed = d.resource_attr_id
                data_to_hide = d.dataset.id

        client.hide_dataset(data_to_hide, [], 'Y', 'Y', 'Y')
        client.user_id = pytest.user_b.id
        netA = client.get_network(network_with_data.id)

        #Find the hidden piece of data and replace it with another
        #to simulate a case of two people working on one attribute
        #where one cannot see the value of it.
        for d in netA.scenarios[0].resourcescenarios:
            if d.resource_attr_id == attr_to_be_changed:
                #THis piece of data is indeed the hidden one.
                assert d.dataset.hidden == 'Y'
                #set the value of the attribute to be a different
                #timeseries.
                dataset = client.JSONObject(dict(
                    type = 'timeseries',
                    name = 'replacement time series',
                    unit = 'feet cubed',
                ))

                t1 = datetime.datetime.now()
                t2 = t1+datetime.timedelta(hours=1)

                ts_val = {0: {t1: [11, 21, 31, 41, 51],
                            t2: [12, 22, 32, 42, 52]}}
                dataset.value = ts_val
                d.dataset = dataset
            else:
                #The rest of the data is unhidden, so should be there.
                assert d.dataset.hidden == 'N'
                assert d.dataset.value is not None

        updated_net = client.update_network(netA)
        updated_net = client.get_network(netA.id)
        #After updating the network, check that the new dataset
        #has been applied
        for d in updated_net.scenarios[0].resourcescenarios:
            if d.resource_attr_id == attr_to_be_changed:
                assert d.dataset.hidden == 'N'
                assert d.dataset.id     != data_to_hide
        #Now validate that the dataset was not overwritten, but replaced
        #by getting the old dataset and ensuring user B can still not see it.
        hidden_dataset = client.get_dataset(data_to_hide)
        assert hidden_dataset.hidden == 'Y'
        assert hidden_dataset.value  == None

        client.user_id = pytest.root_user_id

    def test_edit_hidden_data(self, client, network_with_data):
        """
            test_edit_hidden_data
            Test for the case where one user hides data and another
            user sets the data to something else.

            User A Creates a network with some data
            User A Hides the timeseries created.
            User A shares network with User B

            Check user B cannot see timeseries value
            User B sets value of timeseries to something else.
            Save network.
            Attribute now should have a new, unhidden dataset assigned to that attribute.
        """


        #Let User B view network 1, but not edit it (read_only is 'Y')
        client.share_network(network_with_data.id, ["UserB"], 'N', 'N')

        for d in network_with_data.scenarios[0].resourcescenarios:
            if d.dataset.type == 'timeseries':
                attr_to_be_changed = d.resource_attr_id
                data_to_hide = d.dataset.id
                break

        client.hide_dataset(data_to_hide, [], 'Y', 'Y', 'Y')

        #The next actions are done under user b's identity
        client.user_id = pytest.user_b.id
        netA = client.get_network(network_with_data.id)

        #Find the hidden piece of data and replace it with another
        #to simulate a case of two people working on one attribute
        #where one cannot see the value of it.
        for d in netA.scenarios[0].resourcescenarios:
            if d.resource_attr_id == attr_to_be_changed:
                #THis piece of data is indeed the hidden one.
                assert d.dataset.hidden == 'Y'
                t1 = datetime.datetime.now()
                t2 = t1+datetime.timedelta(hours=1)

                ts_val = {0: {t1: [11, 21, 31, 41, 51],
                            t2: [12, 22, 32, 42, 52]}}
                #Reassign the value of the dataset to something new.
                d.dataset.value = json.dumps(ts_val)
            else:
                #The rest of the data is unhidden, so should be there.
                assert d.dataset.hidden == 'N'
                assert d.dataset.value is not None

        updated_net = client.update_network(netA)
        updated_net = client.get_network(updated_net.id)
        #After updating the network, check that the new dataset
        #has been applied
        for d in  updated_net.scenarios[0].resourcescenarios:
            if d.resource_attr_id == attr_to_be_changed:
                assert d.dataset.hidden == 'N'
                assert d.dataset.id     != data_to_hide
        #Now validate that the dataset was not overwritten, but replaced
        #by getting the old dataset and ensuring user B can still not see it.
        hidden_dataset = client.get_dataset(data_to_hide)
        assert hidden_dataset.hidden == 'Y'
        assert hidden_dataset.value  == None

        #set it back to the root user when we're done
        client.user_id = pytest.root_user_id

class TestiUtilities:
    """
        Test hydra's internal utilities relating to data transformations etc.
    """
    @pytest.mark.parametrize("test_input,expected", [
        ({}, 0),
        ({'a':1}, 1),
        ({'a': {'b': 2}}, 2),
        ({'a': {'b':{}}}, 2),
        ({1: {2: 2}}, 2),
    ])
    def test_count_levels(self, client, test_input, expected):
        assert count_levels(test_input) == expected

    @pytest.mark.parametrize("test_input,expected", [
        ({}, {}),
        ({'a':1}, {'a':1}),
        ({'a': {'b': 2}}, {'a_b': 2}),
        ({'a': {'b':{}}}, {'a_b': {}}),
        ({1: {2: 2}}, {'1_2': 2}),
    ])
    def test_flatten_dict(self, client, test_input, expected):
        assert flatten_dict(test_input) == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import logging
import time
from ..exceptions import HydraError, PermissionError, ResourceNotFoundError
from .. import db
from ..util.permissions import required_perms
from ..db.model import Scenario,\
        ResourceGroupItem,\
        ResourceScenario,\
        TypeAttr,\
        ResourceAttr,\
        Dataset,\
        Metadata,\
        Network,\
        Attr,\
        Node,\
        Link,\
        User,\
        ResourceGroup,\
        ResourceAttrMap

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy import or_, and_, func
from sqlalchemy.orm import joinedload, aliased
from . import data
from collections import namedtuple
from copy import deepcopy

from .network import get_resource

from .objects import JSONObject, Dataset as JSONDataset

log = logging.getLogger(__name__)

def _check_network_ownership(network_id, user_id):
    net = db.DBSession.query(Network).filter(Network.id == network_id).one()

    net.check_write_permission(user_id)

def _get_scenario(scenario_id, user_id, check_write=False, check_can_edit=False):
    log.info("Getting scenario %s", scenario_id)
    try:
        scenario_i = db.DBSession.query(Scenario).filter(Scenario.id == scenario_id).one()

        if check_write is True:
            scenario_i.network.check_write_permission(user_id)
        else:
            #always check read permission. only do this if check write is false to
            #avoid doing basically the same query twice.
            scenario_i.network.check_read_permission(user_id)

        if check_can_edit is True:
            if scenario_i.locked == 'Y':
                raise PermissionError(f'Cannot update scenario {scenario_id} as it is locked.')

        return scenario_i
    except NoResultFound:
        raise ResourceNotFoundError(f"Scenario {scenario_id} does not exist.")

@required_perms("edit_data")
def set_rs_dataset(resource_attr_id, scenario_id, dataset_id, **kwargs):
    rs = db.DBSession.query(ResourceScenario).filter(
        ResourceScenario.resource_attr_id == resource_attr_id,
        ResourceScenario.scenario_id == scenario_id).first()

    if rs is None:
        raise ResourceNotFoundError("Resource scenario for resource attr %s not found in scenario %s"%(resource_attr_id, scenario_id))

    dataset = db.DBSession.query(Dataset).filter(Dataset.id == dataset_id).first()

    if dataset is None:
        raise ResourceNotFoundError("Dataset %s not found"%(dataset_id,))

    rs.dataset_id = dataset_id

    db.DBSession.flush()

    rs = db.DBSession.query(ResourceScenario).filter(
        ResourceScenario.resource_attr_id == resource_attr_id,
        ResourceScenario.scenario_id == scenario_id).first()

    return rs

@required_perms("edit_network")
def merge_scenarios(source_scenario_id, target_scenario_id, match_all_names=True, ignore_missing_attributes=True, **kwargs):
    """
        Merge the data from one scenario to another. Matching is done by node / link name.

        Only attributes which exist on the target scenario are updated. No new attributes are
        added to the target as this would affect other scenarios in that target network which
        is not desireable

        If any node name does nto match, an error is thrown unless overwritten with
        the 'match_all_names = False'
    """
    user_id = kwargs.get('user_id')

    source_scenario = _get_scenario(source_scenario_id, user_id)
    target_scenario = _get_scenario(target_scenario_id, user_id, check_write=True)

    source_network = source_scenario.network
    target_network = target_scenario.network

    source_nodes = {n.name for n in source_network.nodes}
    target_nodes = {n.name for n in target_network.nodes}

    target_resource_lookup = dict((n.name, n) for n in target_network.nodes + target_network.links + target_network.resourcegroups)

    non_matching_node_names = source_nodes - target_nodes
    if len(non_matching_node_names) > 0:
        if match_all_names is True:
            raise HydraError(f"Unable to merge scenario {source_scenario.name} ({source_scenario.id}) into {target_scenario.name} ({target_scenario.id}) as the following node names do not match: {non_matching_node_names}")

        log.warning(f"The following nodes in scenario {source_scenario.name} ({source_scenario.id}) no not match nodes in scenario {target_scenario.name} ({target_scenario.id}) : {non_matching_node_names}")

    cloned_target = clone_scenario(target_scenario.id, user_id=user_id)
    #we need the ORM object here, not a JSONObject
    cloned_scenario = db.DBSession.query(Scenario).filter(Scenario.id == cloned_target.id).one()

    #Make it easy to access resource scenarios from the resource attr ID with a map
    source_ra_rs_map = dict((rs.resource_attr_id, rs) for rs in source_scenario.resourcescenarios)
    target_ra_rs_map = dict((rs.resource_attr_id, rs) for rs in cloned_scenario.resourcescenarios)

    #Create a mapping from the source resource attribute to the target resource
    #attribute so we can then match the resource scenarios
    for source_resource in [source_network] + source_network.nodes + source_network.links + source_network.resourcegroups:

        if hasattr(source_resource, 'projection'):#only a network has a projection
            target_resource = target_network
        else:
            target_resource = target_resource_lookup.get(source_resource.name)

        if target_resource is None:
            log.warning(f"Resource '{source_resource.name}' not found on target. Ignoring.")
            continue

        target_attribute_map = dict((a.attr_id, a.id) for a in target_resource.attributes)
        for source_ra in source_resource.attributes:
            #get the RS of this attribute in both scenarios
            target_ra_id = target_attribute_map.get(source_ra.attr_id)
            if target_ra_id is None:
                if ignore_missing_attributes is True:
                    log.warning(f"Unable to find attribute {source_ra.attr.name} on target {target_resource.name}. Ignoring.")
                    continue
                else:
                    new_ra = ResourceAttr()
                    new_ra.attr_id = source_ra.attr_id
                    new_ra.ref_key = source_ra.ref_key
                    ref = new_ra.ref_key
                    new_ra.node_id = target_resource.id if ref == 'NODE' else None
                    new_ra.network_id = target_resource.id if ref == 'NETWORK' else None
                    new_ra.link_id = target_resource.id if ref == 'LINK' else None
                    new_ra.group_id = target_resource.id if ref == 'GROUP' else None
                    new_ra.attr_is_var = source_ra.attr_is_var
                    db.DBSession.add(new_ra)
                    db.DBSession.flush()
                    target_ra_id = new_ra.id

                    log.info(f"Adding new attribute {source_ra.attr.name} to {target_resource.name}")

            source_rs = source_ra_rs_map.get(source_ra.id)
            target_rs = target_ra_rs_map.get(target_ra_id)
            if source_rs is None:
                continue ## no data associated to the source, so ignore.

            if target_rs is None:
                target_rs = ResourceScenario()
                target_rs.scenario_id = cloned_scenario.id
                target_rs.dataset_id = source_rs.dataset_id
                target_rs.resource_attr_id = target_ra_id
                db.DBSession.add(target_rs)

            #Check if the datasets are different and if so, apply the source dataset ID
            #to the target
            if source_rs.dataset_id != target_rs.dataset_id:
                target_rs.dataset_id = source_rs.dataset_id

    db.DBSession.flush()

@required_perms("edit_network")
def copy_data_from_scenario(resource_attrs, source_scenario_id, target_scenario_id, **kwargs):
    """
        For a given list of resource attribute IDS copy the dataset_ids from
        the resource scenarios in the source scenario to those in the 'target' scenario.
    """

    #Get all the resource scenarios we wish to update
    target_resourcescenarios = db.DBSession.query(ResourceScenario).filter(
        ResourceScenario.scenario_id == target_scenario_id,
        ResourceScenario.resource_attr_id.in_(resource_attrs)).all()

    target_rs_dict = {}
    for target_rs in target_resourcescenarios:
        target_rs_dict[target_rs.resource_attr_id] = target_rs

    #get all the resource scenarios we are using to get our datsets source.
    source_resourcescenarios = db.DBSession.query(ResourceScenario).filter(
        ResourceScenario.scenario_id == source_scenario_id,
        ResourceScenario.resource_attr_id.in_(resource_attrs)).all()

    #If there is an RS in scenario 'source' but not in 'target', then create
    #a new one in 'target'
    for source_rs in source_resourcescenarios:
        target_rs = target_rs_dict.get(source_rs.resource_attr_id)
        if target_rs is not None:
            target_rs.dataset_id = source_rs.dataset_id
        else:
            target_rs = ResourceScenario()
            target_rs.scenario_id = target_scenario_id
            target_rs.dataset_id = source_rs.dataset_id
            target_rs.resource_attr_id = source_rs.resource_attr_id
            db.DBSession.add(target_rs)

    db.DBSession.flush()

    return target_resourcescenarios

@required_perms("get_network")
def get_scenario_by_name(network_id, scenario_name, get_parent_data=False, include_data=True, include_group_items=True, **kwargs):
    """
        Get the specified scenario from a network by its name
        args:
            network_id: The ID of the network to retrieve the scenario from
            scenario_name: The ID of the scenario to retrieve
            get_parent_data: Flag to indicate whether to include the data from the parent scenario also, or just this one.
            include_data: Flag to indicate wheter to return the list of resource scenarios
            include_group_items: Flag to indicate whether to return the list of resource group items
        return:
            A scenario JSONObject
    """

    try:
        scenario_i = db.DBSession.query(Scenario).filter(
            func.lower(Scenario.name) == scenario_name.lower(),
            Scenario.network_id == network_id).first()

        return get_scenario(scenario_i.id,
                            get_parent_data=get_parent_data,
                            include_data=include_data,
                            include_group_items=include_group_items,
                            **kwargs)
    except NoResultFound:
        raise ResourceNotFoundError("Scenario %s not found"%(scenario_name))

@required_perms("get_network")
def get_scenario(scenario_id,
                 get_parent_data=False,
                 include_data=True,
                 include_group_items=True,
                 include_results=True,
                 include_only_results=False,
                 include_metadata=True,
                 include_attr=True,
                 **kwargs):
    """
        Get the specified scenario
        args:
            scenario_id: The ID of the scenario to retrieve
            get_parent_data: Flag to indicate whether to include the data from
                the parent scenario also, or just this one.
            include_data: Flag to indicate wheter to return the list of resource scenarios
            include_group_items: Flag to indicate whether to return the list of resource group items
            include_results: Flag (default True). Set to false if only inputs
                are necessary. This can have a significant performance impact
            include_only_results: Flag (default False). Set to True if you ONLY
                want results data. This will only work when include_results is True
        return:
            A scenario JSONObject
    """

    user_id = kwargs.get('user_id')

    scen_i = _get_scenario(scenario_id, user_id)

    scen_j = JSONObject(scen_i)
    rscen_rs = []
    t = time.time()
    if include_data is True:
        rscen_rs = scen_i.get_data(
            user_id,
            get_parent_data=get_parent_data,
            include_results=include_results,
            include_only_results=include_only_results,
            include_metadata=include_metadata)
    log.info(f"Time taken to get data: {time.time() - t:.2f}")
    rgi_rs = []
    if include_group_items is True:
        rgi_rs = scen_i.get_group_items(get_parent_items=get_parent_data)

    scen_j.resourcescenarios = []

    user = db.DBSession.query(User).filter(User.id == user_id).one()
    is_admin = user.is_admin()

    for rs in rscen_rs:
        rs_j = JSONObject(rs, extras={'resourceattr':JSONObject(rs.resourceattr)})
        scen_j.resourcescenarios.append(rs_j)

    scen_j.resourcegroupitems =[JSONObject(r) for r in rgi_rs]

    return scen_j

@required_perms("edit_network")
def add_scenario(network_id, scenario,**kwargs):
    """
        Add a scenario to a specified network.
    """
    user_id = int(kwargs.get('user_id'))
    log.info("Adding scenarios to network")

    _check_network_ownership(network_id, user_id)

    existing_scen = db.DBSession.query(Scenario).filter(Scenario.name==scenario.name, Scenario.network_id==network_id).first()
    if existing_scen is not None:
        raise HydraError("Scenario with name %s already exists in network %s"%(scenario.name, network_id))

    scen = Scenario()
    scen.name = scenario.name
    scen.description = scenario.description
    scen.layout = scenario.get_layout()
    scen.network_id = network_id
    scen.created_by = user_id
    scen.start_time = scenario.start_time
    scen.end_time = scenario.end_time
    scen.time_step = scenario.time_step
    scen.resourcescenarios = []
    scen.resourcegroupitems = []

    #Just in case someone puts in a negative ID for the scenario.
    if isinstance(scenario.id, int) and scenario.id < 0:
        scenario.id = None

    if scenario.resourcescenarios is not None:
        #extract the data from each resourcescenario so it can all be
        #inserted in one go, rather than one at a time
        all_data = [r.dataset for r in scenario.resourcescenarios]

        datasets = data._bulk_insert_data(all_data, user_id=user_id)

        #record all the resource attribute ids
        resource_attr_ids = [r.resource_attr_id for r in scenario.resourcescenarios]

        #get all the resource scenarios into a list and bulk insert them
        for i, ra_id in enumerate(resource_attr_ids):
            rs_i = ResourceScenario()
            rs_i.resource_attr_id = ra_id
            rs_i.dataset_id = datasets[i].id
            rs_i.scenario_id = scen.id
            rs_i.dataset = datasets[i]
            scen.resourcescenarios.append(rs_i)

    if scenario.resourcegroupitems is not None:
        #Again doing bulk insert.
        for group_item in scenario.resourcegroupitems:
            group_item_i = ResourceGroupItem()
            group_item_i.scenario_id = scen.id
            group_item_i.group_id = group_item.group_id
            group_item_i.ref_key = group_item.ref_key
            if group_item.ref_key == 'NODE':
                group_item_i.node_id = group_item.ref_id
            elif group_item.ref_key == 'LINK':
                group_item_i.link_id = group_item.ref_id
            elif group_item.ref_key == 'GROUP':
                group_item_i.subgroup_id = group_item.ref_id
            scen.resourcegroupitems.append(group_item_i)
    db.DBSession.add(scen)
    db.DBSession.flush()
    return scen

@required_perms("edit_network", "edit_data")
def update_scenario(scenario, update_data=True, update_groups=True, flush=True, **kwargs):
    """
        Update a single scenario
        as all resources already exist, there is no need to worry
        about negative IDS

        flush = True flushes to the DB at the end of the function.
        flush = False does not flush, assuming that it will happen as part
                of another process, like update_network.
    """
    user_id = kwargs.get('user_id')
    scen = _get_scenario(scenario.id, user_id)

    if scen.locked == 'Y':
        raise PermissionError('Scenario is locked. Unlock before editing.')

    scen.name = scenario.name
    scen.description = scenario.description
    scen.layout = scenario.get_layout()
    scen.start_time = scenario.start_time
    scen.end_time = scenario.end_time
    scen.time_step = scenario.time_step

    if scenario.resourcescenarios == None:
        scenario.resourcescenarios = []
    if scenario.resourcegroupitems == None:
        scenario.resourcegroupitems = []

    #lazy load resourcescenarios from the DB
    scen.resourcescenarios

    #creat a reverse mapping from the resource_attr_id to the RS to avoid querying
    #the table again later.
    #no need to use scenario id in key here cos it's always in the context of one scenario
    rsmap = {}
    for rs in scen.resourcescenarios:
        rsmap[rs.resource_attr_id] = rs

    if update_data is True:
        datasets = [rs.dataset for rs in scenario.resourcescenarios]
        updated_datasets = data._bulk_insert_data(datasets, user_id, kwargs.get('app_name'))
        for i, r_scen in enumerate(scenario.resourcescenarios):
            log.debug("updating resource scenario...")

            rscen_i = rsmap.get(r_scen.resource_attr_id)
            _update_resourcescenario(scen, r_scen, r_scen_i=rscen_i, dataset=updated_datasets[i], user_id=user_id, source=kwargs.get('app_name'))

    log.info('%s Resource Scenarios Updated', len(scenario.resourcescenarios))

    #lazy load resource grou items from the DB
    scen.resourcegroupitems

    if update_groups is True:
        #Get all the exiting resource group items for this scenario.
        #THen process all the items sent to this handler.
        #Any in the DB that are not passed in here are removed.
        for group_item in scenario.resourcegroupitems:
            _add_resourcegroupitem(group_item, scenario.id)

    log.info('Resource Group Items Updated')

    if flush is True:
        db.DBSession.flush()

    log.info('Scenario %s updated', scenario.name)

    return scen

@required_perms("edit_network")
def set_scenario_status(scenario_id, status, **kwargs):
    """
        Set the status of a scenario.
    """

    user_id = kwargs.get('user_id')

    scenario_i = _get_scenario(scenario_id, user_id, check_write=True, check_can_edit=True)

    scenario_i.status = status
    db.DBSession.flush()
    return 'OK'

@required_perms("edit_network")
def purge_scenario(scenario_id, **kwargs):
    """
        Set the status of a scenario.
    """

    user_id = kwargs.get('user_id')

    scenario_i = _get_scenario(scenario_id, user_id, check_write=True, check_can_edit=True)

    db.DBSession.delete(scenario_i)
    db.DBSession.flush()
    return 'OK'

@required_perms("edit_network")
def create_child_scenario(parent_scenario_id, child_name, **kwargs):
    """
        Create a new scenario which inherits from the given scenario. The new
        scenario contains no resource scenarios or groups. All data is inherited
        from the parent.
    """

    user_id = kwargs.get('user_id')

    scen_i = _get_scenario(parent_scenario_id, user_id)

    log.info("Creating child scenario of %s", scen_i.name)

    existing_scenarios = db.DBSession.query(Scenario).filter(Scenario.network_id==scen_i.network_id).all()
    if child_name is not None:
        for existing_scenario in existing_scenarios:
            if existing_scenario.name == child_name:
                raise HydraError("A scenario with the name {0} already exists in this network. ".format(child_name))
    else:
        child_name = "%s (child)"%(scen_i.name)
        num_child_scenarios = 0
        for existing_scenario in existing_scenarios:
            if existing_scenario.name.find(child_name) >= 0:
                num_child_scenarios = num_child_scenarios + 1

        if num_child_scenarios > 0:
            child_name = child_name + " %s"%(num_child_scenarios)


    child_scen = Scenario()
    child_scen.network_id           = scen_i.network_id
    child_scen.name                 = child_name
    child_scen.description          = scen_i.description
    child_scen.created_by           = kwargs['user_id']
    child_scen.parent_id            = scen_i.id

    child_scen.start_time           = scen_i.start_time
    child_scen.end_time             = scen_i.end_time
    child_scen.time_step            = scen_i.time_step

    db.DBSession.add(child_scen)

    db.DBSession.flush()

    log.info("New scenario created. Scenario ID: %s", child_scen.id)

    child_scenario_j = JSONObject(_get_scenario(child_scen.id, user_id))
    child_scenario_j.resourcescenarios = []
    child_scenario_j.resourcegroupitems =[]
    log.info("Returning child scenario")

    return child_scenario_j

@required_perms("edit_network")
def clone_scenario(scenario_id, retain_results=False, scenario_name=None, **kwargs):
    """
        Create an exact copy of a scenario and place it in the same network
        args:
            scenario_id (int): The scenario ID to clone
            retain_results (bool): Flag to indicated whether resource scenarios connected to resource attribtues which have an 'attr_is_var' are copied or not. Defaults to False, so results are not retained by default.
            scenario_name (string): The name of the new scenario. If None, the existing scenario's name is used, appended with '(clone'). Multiple clones of the same network result in  "... (clone) 1", "... (clone) 2" etc,
    """

    user_id = kwargs.get('user_id')

    scen_i = _get_scenario(scenario_id, user_id)

    log.info("cloning scenario %s", scen_i.name)


    if scenario_name is None:
        existing_scenarios = db.DBSession.query(Scenario).filter(Scenario.network_id==scen_i.network_id).all()
        num_cloned_scenarios = 0
        for existing_sceanrio in existing_scenarios:
            if existing_sceanrio.name.find('clone') >= 0:
                num_cloned_scenarios = num_cloned_scenarios + 1

        cloned_name = "%s (clone)"%(scen_i.name)
        if num_cloned_scenarios > 0:
            cloned_name = cloned_name + " %s"%(num_cloned_scenarios)
    else:
        cloned_name = scenario_name

        #check this scenario name is available
        existing_scenario_with_name = db.DBSession.query(Scenario).filter(
            Scenario.network_id==scen_i.network_id, Scenario.name==scenario_name).first()

        if existing_scenario_with_name is not None:
            netname = existing_scenario_with_name.network.name
            raise HydraError(f"A scenario with name {scenario_name} already exists in network {netname}")

    log.info("Cloned scenario name is %s", cloned_name)

    cloned_scen = Scenario()
    cloned_scen.network_id           = scen_i.network_id
    cloned_scen.name                 = cloned_name
    cloned_scen.description          = scen_i.description
    cloned_scen.created_by           = kwargs['user_id']
    cloned_scen.parent_id            = scen_i.parent_id

    cloned_scen.start_time           = scen_i.start_time
    cloned_scen.end_time             = scen_i.end_time
    cloned_scen.time_step            = scen_i.time_step

    db.DBSession.add(cloned_scen)

    db.DBSession.flush()

    cloned_scenario_id = cloned_scen.id
    log.info("New scenario created. Scenario ID: %s", cloned_scenario_id)


    log.info("Getting in resource scenarios to clone from scenario %s", scenario_id)
    if retain_results is False:
        old_rscen_rs = db.DBSession.query(ResourceScenario).filter(
            ResourceScenario.scenario_id == scenario_id,
            ResourceAttr.id == ResourceScenario.resource_attr_id,
            ResourceAttr.attr_is_var == 'N'
        ).all()
    else:
        old_rscen_rs = db.DBSession.query(ResourceScenario).filter(
            ResourceScenario.scenario_id == scenario_id
        ).all()

    new_rscens = []
    for old_rscen in old_rscen_rs:
        new_rscens.append(dict(
            dataset_id=old_rscen.dataset_id,
            scenario_id=cloned_scenario_id,
            resource_attr_id=old_rscen.resource_attr_id,
            source = kwargs.get('app_name', old_rscen.source)
        ))

    if len(new_rscens) > 0:
        db.DBSession.execute(ResourceScenario.__table__.insert(), new_rscens)

    log.info("ResourceScenarios cloned")

    log.info("Getting old resource group items for scenario %s", scenario_id)
    old_rgis = db.DBSession.query(ResourceGroupItem).filter(
        ResourceGroupItem.scenario_id == scenario_id).all()
    new_rgis = []
    for old_rgi in old_rgis:
        new_rgis.append(dict(
            ref_key=old_rgi.ref_key,
            node_id=old_rgi.node_id,
            link_id=old_rgi.link_id,
            subgroup_id=old_rgi.subgroup_id,
            group_id=old_rgi.group_id,
            scenario_id=cloned_scenario_id,
        ))
    if len(new_rgis) > 0:
        db.DBSession.execute(ResourceGroupItem.__table__.insert(), new_rgis)

    log.info("Cloning finished.")

    log.info("Retrieving cloned scenario")
    new_rscen_rs = db.DBSession.query(ResourceScenario).filter(
        ResourceScenario.scenario_id == cloned_scenario_id).all()
    new_rgis_rs  = db.DBSession.query(ResourceGroupItem).filter(
        ResourceGroupItem.scenario_id == cloned_scenario_id).all()

    cloned_scen = JSONObject(_get_scenario(cloned_scenario_id, user_id))
    cloned_scen.resourcescenarios = [JSONObject(r) for r in new_rscen_rs]
    cloned_scen.resourcegroupitems =[JSONObject(r) for r in new_rgis_rs]
    log.info("Returning cloned scenario")


    return cloned_scen

def _get_dataset_as_dict(rs, user_id):
    if rs.dataset is None:
        return None

    dataset = JSONDataset(rs.dataset)

    dataset['metadata'] = {}

    try:
        rs.dataset.check_read_permission(user_id)
    except PermissionError:
           dataset['value']      = None
           dataset['metadata'] = {}

    return dataset

def _get_scenario_network_resources(scenario_id, scenario_i=None):
    """
        Inspect the network of the given scenario and return a mapping of ID to Name
        for each element in that network, categorised by resource type.
    """
    if scenario_i is None:
        scenario_i = db.DBSession.query(Scenario).filter(Scenario.id==scenario_id).one()

    net = db.DBSession.query(Network.id, Network.name).filter(Network.id == scenario_i.network_id).all()
    nodes = db.DBSession.query(Node.id, Node.name).filter(Node.network_id == scenario_i.network_id).all()
    links = db.DBSession.query(Link.id, Link.name).filter(Link.network_id == scenario_i.network_id).all()
    groups = db.DBSession.query(ResourceGroup.id, ResourceGroup.name).filter(ResourceGroup.network_id == scenario_i.network_id).all()

    mapping_dict = {}

    mapping_dict['NETWORK'] = dict(net)
    mapping_dict['NODE'] = dict(nodes)
    mapping_dict['LINK'] = dict(links)
    mapping_dict['GROUP'] = dict(groups)

    return mapping_dict

def _get_resource_id(resourceattr):
    for col in ('node_id', 'link_id', 'group_id', 'network_id'):
        if resourceattr.get(col) is not None:
            return resourceattr.get(col)

@required_perms("get_network")
def compare_scenarios(scenario_id_1, scenario_id_2, allow_different_networks=False, include_results=False, **kwargs):
    """
        Compare two scenarios and return a 'diff' dictionary containing all the differences.
        Args:
            scenario_1_id (int): Scenario 1 ID
            scenario_2_id (int): Scenario 2 ID
            allow_different_networks (bool) (default False); Flag to indicate whether it should be allowed to compare the scenarios from two independent networks.
            include_results (bool) (default False): If set to true, includes all 'attr_is_var' values. Otherwise it only compares inputs.
        returns:
            dict: Containing the differences
        raises:
            HydraError if the scenarios are not in the same network (and the allow_different_networks is false)
    """
    user_id = kwargs.get('user_id')

    scenario_1 = get_scenario(scenario_id_1,
                              include_parent_data=True,
                              include_results=include_results,
                              include_attr=True,
                              include_metadata=False,
                              user_id=user_id)

    scenario_2 = get_scenario(scenario_id_2,
                              include_parent_data=True,
                              include_results=include_results,
                              include_attr=True,
                              include_metadata=False,
                              user_id=user_id)

    if allow_different_networks is False and scenario_1.network_id != scenario_2.network_id:
        raise HydraError("Cannot compare scenarios that are not"
                         " in the same network!")

    #find a mapping from ID to Name for all nodes / links / groups in scenario 1's network
    s1_resource_mapping = _get_scenario_network_resources(scenario_1.id, scenario_1)
    if scenario_1.network_id == scenario_2.network_id:
        s2_resource_mapping = s1_resource_mapping
    else:
        s2_resource_mapping = _get_scenario_network_resources(scenario_2.id, scenario_2)


    scenario_1_rs = scenario_1.resourcescenarios
    scenario_2_rs = scenario_2.resourcescenarios

    scenario_1_rgi = scenario_1.resourcegroupitems
    scenario_2_rgi = scenario_2.resourcegroupitems


    scenariodiff = dict(
       object_type = 'ScenarioDiff'
    )
    resource_diffs = []

    #Make a list of all the resource scenarios (aka data) that are unique
    #to scenario 1 and that are in both scenarios, but are not the same.

    #For efficiency, build a dictionary of the data in scenarios and refer
    #them rather than nesting for loops.
    r_scen_1_dict = dict()
    r_scen_2_dict = dict()
    for s1_rs in scenario_1_rs:
        r_scen_1_dict[s1_rs.resource_attr_id] = s1_rs
    for s2_rs in scenario_2_rs:
        r_scen_2_dict[s2_rs.resource_attr_id] = s2_rs

    rscen_1_dataset_ids = set([r_scen.dataset_id for r_scen in scenario_1_rs])
    rscen_2_dataset_ids = set([r_scen.dataset_id for r_scen in scenario_2_rs])

    log.info("Datasets In 1 not in 2: %s"%(len(rscen_1_dataset_ids - rscen_2_dataset_ids)))
    log.info("Datasets In 2 not in 1: %s"%(len(rscen_2_dataset_ids - rscen_1_dataset_ids)))

    for ra_id, s1_rs in r_scen_1_dict.items():
        s2_rs = r_scen_2_dict.get(ra_id)
        if s2_rs is not None:
            log.debug("Is %s == %s?"%(s1_rs.dataset_id, s2_rs.dataset_id))
            if s1_rs.dataset_id != s2_rs.dataset_id:
                resource_diff = dict(
                    resource_attr_id = s1_rs.resource_attr_id,
                    scenario_1_dataset = s1_rs.dataset,
                    scenario_2_dataset = s2_rs.dataset,
                    attr_name = s1_rs.resourceattr.attr.name,
                    resource_name = s1_resource_mapping[s1_rs.resourceattr.ref_key][_get_resource_id(s1_rs.resourceattr)],
                )
                resource_diffs.append(resource_diff)

            continue
        else:
            #this is unique in scenario 1
            resource_diff = dict(
                resource_attr_id = s1_rs.resource_attr_id,
                scenario_1_dataset = s1_rs.dataset,
                scenario_2_dataset = None,
                attr_name = s1_rs.resourceattr.attr.name,
                resource_name = s1_resource_mapping[s1_rs.resourceattr.ref_key][_get_resource_id(s1_rs.resourceattr)],
            )
            resource_diffs.append(resource_diff)

    #make a list of all the resource scenarios (aka data) that are unique
    #in scenario 2.
    for ra_id, s2_rs in r_scen_2_dict.items():
        s1_rs = r_scen_1_dict.get(ra_id)
        if s1_rs is None:
            resource_diff = dict(
                resource_attr_id = ra_id,
                scenario_1_dataset = None,
                scenario_2_dataset = s2_rs.dataset,
                attr_name = s2_rs.resourceattr.attr.name,
                resource_name = s2_resource_mapping[s2_rs.resourceattr.ref_key][_get_resource_id(s2_rs.resourceattr)],
            )
            resource_diffs.append(resource_diff)

    scenariodiff['resourcescenarios'] = resource_diffs

    #Now compare groups.
    #Return list of group items in scenario 1 not in scenario 2 and vice versa
    s1_items = []
    for s1_item in scenario_1_rgi:
        s1_items.append((s1_item.group_id, s1_item.ref_key, s1_item.node_id, s1_item.link_id, s1_item.subgroup_id))
    s2_items = []
    for s2_item in scenario_2_rgi:
        s2_items.append((s2_item.group_id, s2_item.ref_key, s2_item.node_id, s2_item.link_id, s2_item.subgroup_id))

    groupdiff = dict()
    scenario_1_items = []
    scenario_2_items = []
    for s1_only_item in set(s1_items) - set(s2_items):

        item = ResourceGroupItem(
            group_id = s1_only_item[0],
            ref_key  = s1_only_item[1],
            node_id   = s1_only_item[2],
            link_id   = s1_only_item[3],
            subgroup_id   = s1_only_item[4],
        )
        scenario_1_items.append(item)
    for s2_only_item in set(s2_items) - set(s1_items):
        item = ResourceGroupItem(
            group_id = s2_only_item[0],
            ref_key  = s2_only_item[1],
            node_id   = s2_only_item[2],
            link_id   = s2_only_item[3],
            subgroup_id   = s2_only_item[4],
        )
        scenario_2_items.append(item)

    groupdiff['scenario_1_items'] = scenario_1_items
    groupdiff['scenario_2_items'] = scenario_2_items
    scenariodiff['groups'] = groupdiff

    return scenariodiff

@required_perms("get_data")
def get_resource_scenario(resource_attr_id, scenario_id, get_parent_data=False, **kwargs):
    """
        Get the resource scenario object for a given resource atttribute and scenario.
        This is done when you know the attribute, resource and scenario and want to get the
        value associated with it.

        The get_parent_data flag indicates whether we should look only at this scenario, or if
        the resource scenario does not exist on this scenario to look in its parent.
    """
    user_id = kwargs.get('user_id')

    scenario_i = _get_scenario(scenario_id, user_id)

    scenario_rs = scenario_i.get_data(user_id, get_parent_data=get_parent_data, ra_ids=[resource_attr_id])

    for rs_i in scenario_rs:
        if rs_i.resource_attr_id == resource_attr_id:
            return rs_i
    else:
        raise ResourceNotFoundError("resource scenario for %s not found in scenario %s"%(resource_attr_id, scenario_id))

@required_perms("get_data")
def get_resourceattr_data(resource_attr_ids, scenario_id, get_parent_data=False, **kwargs):
    """
        Get the resource scenarios associated to a list of resource attribute IDs
    """
    user_id = kwargs.get('user_id')

    scenario_i = _get_scenario(scenario_id, user_id)

    if not isinstance(resource_attr_ids, list):
        resource_attr_ids = [resource_attr_ids]

    scenario_rs = scenario_i.get_data(user_id, get_parent_data=get_parent_data, ra_ids=resource_attr_ids)


    resource_scenario_dict = {}

    for rs_i in scenario_rs:
        if rs_i.resource_attr_id in resource_attr_ids:
            rs_i.dataset
            rs_i.dataset.metadata
            resource_scenario_dict[rs_i.resource_attr_id] = rs_i

    return resource_scenario_dict

@required_perms("edit_network")
def lock_scenario(scenario_id, **kwargs):
    """
        Unlock a scenario
    """
    #user_id = kwargs.get('user_id')
    #check_perm(user_id, 'edit_network')

    user_id = kwargs.get('user_id')

    scenario_i = _get_scenario(scenario_id, user_id)

    scenario_i.network.check_write_permission(user_id)

    scenario_i.locked = 'Y'

    db.DBSession.flush()
    return 'OK'

@required_perms("edit_network")
def unlock_scenario(scenario_id, **kwargs):
    """
        Unlock a scenario
    """
    #user_id = kwargs.get('user_id')
    #check_perm(user_id, 'edit_network')
    user_id = kwargs.get('user_id')

    scenario_i = _get_scenario(scenario_id, user_id)

    scenario_i.network.check_write_permission(user_id)

    scenario_i.locked = 'N'

    db.DBSession.flush()
    return 'OK'

@required_perms("get_data")
def get_dataset_scenarios(dataset_id, **kwargs):

    try:
        db.DBSession.query(Dataset).filter(Dataset.id == dataset_id).one()
    except NoResultFound:
        raise ResourceNotFoundError("Dataset %s not found"%dataset_id)

    log.info("dataset %s exists", dataset_id)

    scenarios = db.DBSession.query(Scenario).filter(
        Scenario.status == 'A',
        ResourceScenario.scenario_id == Scenario.id,
        ResourceScenario.dataset_id == dataset_id).distinct().all()

    log.info("%s scenarios retrieved", len(scenarios))

    return scenarios

@required_perms("edit_data", "edit_network")
def bulk_update_resourcedata(scenario_ids, resource_scenarios, **kwargs):
    """
        Update the data associated with a list of scenarios.
    """
    user_id = kwargs.get('user_id')
    res = None

    res = {}

    net_ids = db.DBSession.query(Scenario.network_id).filter(Scenario.id.in_(scenario_ids)).all()

    if len(set(net_ids)) != 1:
        raise HydraError("Scenario IDS are not in the same network")

    for scenario_id in scenario_ids:

        scen_i = _get_scenario(scenario_id, user_id, check_write=True, check_can_edit=True)

        #this is cast as a string so it can be read into a JSONObject
        res[str(scenario_id)] = []

        #make a lookup dict of all the resource scenarios that already exist from the
        #ones that have been passed in to avoid querying for every one individually.

        ra_ids = [rs.resource_attr_id for rs in resource_scenarios]
        r_scens_i = db.DBSession.query(ResourceScenario).filter(
                ResourceScenario.scenario_id == scenario_id,
                ResourceScenario.resource_attr_id.in_(ra_ids)).all()
        r_scen_dict = dict((rs.resource_attr_id, rs) for rs in r_scens_i)
        for rs in resource_scenarios:
            if rs.dataset is not None:
                updated_rs = _update_resourcescenario(scen_i,
                                                      rs,
                                                      r_scen_i=r_scen_dict.get(rs.resource_attr_id),
                                                      user_id=user_id,
                                                      source=kwargs.get('app_name'))
                #this is cast as a string so it can be read into a JSONObject
                res[str(scenario_id)].append(updated_rs)
            else:
                _delete_resourcescenario(scenario_id, rs.resource_attr_id)

        db.DBSession.flush()

    return res

@required_perms("edit_data", "edit_network")
def update_resourcedata(scenario_id, resource_scenarios, **kwargs):
    """
        Update the data associated with a scenario.
        Data missing from the resource scenario will not be removed
        from the scenario. Use the remove_resourcedata for this task.

        If the resource scenario does not exist, it will be created.
        If the value of the resource scenario is specified as being None, the
        resource scenario will be deleted.
        If the value of the resource scenario does not exist, it will be created.
        If the both the resource scenario and value already exist, the resource scenario
        will be updated with the ID of the dataset.

        If the dataset being set is being changed, already exists,
        and is only used by a single resource scenario,
        then the dataset itself is updated, rather than a new one being created.
    """
    user_id = kwargs.get('user_id')
    res = None

    scenario_i = _get_scenario(scenario_id, user_id, check_write=True, check_can_edit=True)

    res = []
    for rs in resource_scenarios:
        if rs.dataset is not None:
            updated_rs = _update_resourcescenario(scenario_i, rs, user_id=user_id, source=kwargs.get('app_name'))
            res.append(updated_rs)
        else:
            _delete_resourcescenario(scenario_id, rs.resource_attr_id)

    db.DBSession.flush()

    return res

@required_perms("edit_data", "edit_network")
def delete_scenario_results(scenario_id, **kwargs):
    """
        Delete all the resource scenarios in a scenario which are linked to
        resource attributes that have 'attr_is_var' set to 'Y'
    """
    user_id = kwargs.get('user_id')

    _get_scenario(scenario_id, user_id, check_write=True, check_can_edit=True)

    results_rs = db.DBSession.query(ResourceScenario)\
        .join(ResourceAttr)\
        .filter(ResourceScenario.scenario_id == scenario_id)\
        .filter(ResourceAttr.attr_is_var == 'Y').all()
    for rs in results_rs:
        db.DBSession.delete(rs)

    db.DBSession.flush()

    log.info("%s resource scenarios deleted", len(results_rs))

@required_perms("edit_data", "edit_network")
def delete_resource_scenario(scenario_id, resource_attr_id, quiet=False, **kwargs):
    """
        Remove the data associated with a resource in a scenario.
    """
    user_id = kwargs.get('user_id')

    _get_scenario(scenario_id, user_id, check_write=True, check_can_edit=True)

    _delete_resourcescenario(scenario_id, resource_attr_id, suppress_error=quiet)

@required_perms("edit_data", "edit_network")
def delete_resource_scenarios(scenario_id, resource_attr_ids, quiet=False, **kwargs):
    """
        Remove the data associated with a list of resources in a scenario.
    """
    user_id = kwargs.get('user_id')

    log.info("Deleting %s resource scenarios from from scenario %s", len(resource_attr_ids), scenario_id)
    _get_scenario(scenario_id, user_id, check_write=True, check_can_edit=True)

    for resource_attr_id in resource_attr_ids:
        _delete_resourcescenario(scenario_id, resource_attr_id, suppress_error=quiet)

@required_perms("edit_data", "edit_network")
def delete_resourcedata(scenario_id, resource_scenario, quiet = False, **kwargs):
    """
        Remove the data associated with a resource in a scenario.
        The 'quiet' parameter indicates whether an non-existent RS should throw
        an error.
    """
    user_id = kwargs.get('user_id')

    _get_scenario(scenario_id, user_id, check_write=True, check_can_edit=True)

    _delete_resourcescenario(scenario_id, resource_scenario.resource_attr_id, suppress_error=quiet)


def _delete_resourcescenario(scenario_id, resource_attr_id, suppress_error=False):

    log.debug("Deleting resource scenario for RA %s from scenario %s", resource_attr_id, scenario_id)

    try:
        sd_i = db.DBSession.query(ResourceScenario).filter(
            ResourceScenario.scenario_id == scenario_id,
            ResourceScenario.resource_attr_id == resource_attr_id).one()
    except NoResultFound:
        if suppress_error is False:
            raise HydraError("ResourceAttr %s does not exist in scenario %s."%(resource_attr_id, scenario_id))
        return

    db.DBSession.delete(sd_i)
    db.DBSession.flush()

def _update_resourcescenario(scenario, resource_scenario, r_scen_i=None, dataset=None, new=False, user_id=None, source=None):
    """
        Insert or Update the value of a resource's attribute by first getting the
        resource, then parsing the input data, then assigning the value.

        returns a ResourceScenario object.
    """
    if scenario is None:
        scenario = db.DBSession.query(Scenario).filter(Scenario.id == 1).one()

    ra_id = resource_scenario.resource_attr_id

    log.debug("Assigning resource attribute: %s",ra_id)
    #count the number of new RS to report it in the logs
    new_rs = 0
    if r_scen_i is None:
        try:
            r_scen_i = db.DBSession.query(ResourceScenario).filter(
                ResourceScenario.scenario_id == scenario.id,
                ResourceScenario.resource_attr_id == resource_scenario.resource_attr_id).one()
        except NoResultFound as e:
            log.debug("Creating new RS for RS %s in scenario %s", resource_scenario.resource_attr_id, scenario.id)
            r_scen_i = ResourceScenario()
            r_scen_i.resource_attr_id = resource_scenario.resource_attr_id
            r_scen_i.scenario_id = scenario.id
            r_scen_i.scenario = scenario
            new_rs = new_rs + 1

            db.DBSession.add(r_scen_i)

    if scenario.locked == 'Y':
        log.info("Scenario %s is locked", scenario.id)
        return r_scen_i


    if dataset is not None:
        r_scen_i.dataset = dataset

        return r_scen_i

    dataset = resource_scenario.dataset

    dataset_j = JSONDataset(dataset)

    value = dataset_j.parse_value()

    log.debug("Assigning %s to resource attribute: %s", value, ra_id)

    if value is None:
        log.info("Cannot set data on resource attribute %s", ra_id)
        return None

    metadata = dataset_j.get_metadata_as_dict(source=source, user_id=user_id)
    data_unit_id = dataset_j.unit_id

    data_hash = dataset_j.get_hash(value, metadata)

    new_rscen_i = assign_value(r_scen_i,
                 dataset_j.type.lower(),
                 value,
                 data_unit_id,
                 dataset_j.name,
                 metadata=metadata,
                 data_hash=data_hash,
                 user_id=user_id,
                 source=source)

    return new_rscen_i

@required_perms("edit_data", "edit_network")
def assign_value(rs, data_type, val,
                 unit_id, name, metadata={}, data_hash=None, user_id=None, source=None):
    """
        Insert or update a piece of data in a scenario.
        If the dataset is being shared by other resource scenarios, a new dataset is inserted.
        If the dataset is ONLY being used by the resource scenario in question, the dataset
        is updated to avoid unnecessary duplication.
    """

    log.debug("Assigning value %s to rs %s in scenario %s",
              name, rs.resource_attr_id, rs.scenario_id)

    if rs.scenario.locked == 'Y':
        raise PermissionError("Cannot assign value. Scenario %s is locked"
                              %(rs.scenario_id))

    #Check if this RS is the only RS in the DB connected to this dataset.
    #If no results is found, the RS isn't in the DB yet, so the condition is false.
    update_dataset = False # Default behaviour is to create a new dataset.

    if rs.dataset is not None:

        #Has this dataset changed?
        if rs.dataset.hash == data_hash:
            log.info("Dataset has not changed. Returning.")
            return rs

        connected_rs = db.DBSession.query(ResourceScenario).filter(ResourceScenario.dataset_id == rs.dataset.id).all()
        #If there's no RS found, then the incoming rs is new, so the dataset can be altered
        #without fear of affecting something else.
        if len(connected_rs) == 0:
        #If it's 1, the RS exists in the DB, but it's the only one using this dataset or
        #The RS isn't in the DB yet and the datset is being used by 1 other RS.
            update_dataset = True

        if len(connected_rs) == 1:
            if connected_rs[0].scenario_id == rs.scenario_id and connected_rs[0].resource_attr_id == rs.resource_attr_id:
                update_dataset = True
        else:
            update_dataset = False

    if update_dataset is True:
        log.info("Updating dataset '%s'", name)
        dataset = data.update_dataset(rs.dataset.id, name, data_type, val, unit_id, metadata, flush=False, **dict(user_id=user_id))
        log.info("Updated dataset '%s'", name)
        rs.dataset = dataset
        rs.dataset_id = dataset.id
        log.info("Set RS dataset id to %s"%dataset.id)
    else:
        log.info("Creating new dataset %s in scenario %s", name, rs.scenario_id)
        dataset = data.add_dataset(
            data_type,
            val,
            unit_id,
            metadata=metadata,
            name=name,
            **dict(user_id=user_id)
        )
        rs.dataset = dataset
        rs.source = source


    db.DBSession.flush()

    return rs

@required_perms("edit_data", "edit_network")
def add_data_to_attribute(scenario_id, resource_attr_id, dataset,**kwargs):
    """
        Add data to a resource scenario outside of a network update
    """
    user_id = kwargs.get('user_id')

    scenario_i = _get_scenario(scenario_id, user_id, check_write=True, check_can_edit=True)

    try:
        r_scen_i = db.DBSession.query(ResourceScenario).filter(
            ResourceScenario.scenario_id == scenario_id,
            ResourceScenario.resource_attr_id == resource_attr_id).one()
        log.info("Existing resource scenario found for %s in scenario %s", resource_attr_id, scenario_id)
    except NoResultFound:
        log.info("No existing resource scenarios found for %s in scenario %s. Adding a new one.", resource_attr_id, scenario_id)
        r_scen_i = ResourceScenario()
        r_scen_i.scenario_id      = scenario_id
        r_scen_i.resource_attr_id = resource_attr_id
        scenario_i.resourcescenarios.append(r_scen_i)

    data_type = dataset.type.lower()

    dataset_j = JSONDataset(dataset)
    value = dataset_j.parse_value()

    dataset_metadata = dataset_j.get_metadata_as_dict(user_id=kwargs.get('user_id'),
                                                      source=kwargs.get('source'))
    if value is None:
        raise HydraError(f"Cannot set value to attribute. No value was sent with dataset {dataset_j.id}")

    data_hash = dataset_j.get_hash(value, dataset_metadata)

    new_rscen_i = assign_value(r_scen_i, data_type, value, dataset_j.unit_id, dataset_j.name,
                          metadata=dataset_metadata, data_hash=data_hash, user_id=user_id)

    db.DBSession.flush()

    return new_rscen_i

@required_perms("get_data", "get_network")
def get_scenario_data(scenario_id, get_parent_data=False, **kwargs):
    """
        Get all the datasets from the scenario with the specified ID
        @returns a list of dictionaries
    """
    user_id = kwargs.get('user_id')

    scenario_i = _get_scenario(scenario_id, user_id)

    scenario_rs = scenario_i.get_data(user_id, get_parent_data=get_parent_data)

    dataset_ids = []
    datasets = []
    for rs in scenario_rs:
        if rs.dataset.id in dataset_ids:
            continue

        if rs.dataset.hidden == 'Y':
            try:
                rs.dataset.check_read_permission(user_id)
            except:
                rs.dataset.value = None
                rs.dataset.metadata = []

        datasets.append(rs.dataset)
        dataset_ids.append(rs.dataset.id)

    log.info("Retrieved %s datasets", len(datasets))
    return datasets

@required_perms("get_data", "get_network")
def get_attribute_data(attr_ids, node_ids, **kwargs):
    """
        For a given attribute or set of attributes, return  all the resources and
        resource scenarios in the network
    """
    node_attrs = db.DBSession.query(ResourceAttr).\
        options(joinedload(ResourceAttr.attr)).\
        filter(ResourceAttr.node_id.in_(node_ids),
               ResourceAttr.attr_id.in_(attr_ids)).all()

    ra_ids = []
    for ra in node_attrs:
        ra_ids.append(ra.id)


    resource_scenarios = db.DBSession.query(ResourceScenario).filter(
        ResourceScenario.resource_attr_id.in_(ra_ids)).options(
            joinedload(ResourceScenario.resourceattr)).options(
                joinedload(ResourceScenario.dataset).joinedload(Dataset.metadata)
            ).order_by(ResourceScenario.scenario_id).all()


    for rs in resource_scenarios:
        if rs.dataset.hidden == 'Y':
            try:
                rs.dataset.check_read_permission(kwargs.get('user_id'))
            except:
                rs.dataset.value = None
        db.DBSession.expunge(rs)

    return node_attrs, resource_scenarios

def _get_all_network_resource_attributes(network_id):
    """
        Get all the attributes for the nodes, links and groups of a network.
        Return these attributes as a dictionary, keyed on type (NODE, LINK, GROUP)
        then by ID of the node or link.

        args:
            network_id (int) The ID of the network from which to retrieve the attributes
        returns:
            A list of sqlalchemy result proxy objects
    """
    base_qry = db.DBSession.query(ResourceAttr).filter(Attr.id==ResourceAttr.attr_id)

    all_node_attribute_qry = base_qry.join(Node).filter(Node.network_id == network_id)

    all_link_attribute_qry = base_qry.join(Link).filter(Link.network_id == network_id)

    all_group_attribute_qry = base_qry.join(ResourceGroup)\
            .filter(ResourceGroup.network_id == network_id)

    network_attribute_qry = base_qry.filter(ResourceAttr.network_id == network_id)

    logging.info("Getting all attributes using execute")
    attribute_qry = all_node_attribute_qry.union(all_link_attribute_qry,
                                                 all_group_attribute_qry,
                                                 network_attribute_qry)
    all_resource_attributes = attribute_qry.all()

    return all_resource_attributes

@required_perms("get_data", "get_network")
def get_resource_data(ref_key,
                      ref_id,
                      scenario_id,
                      type_id=None,
                      expunge_session=True,
                      get_parent_data=False,
                      include_inputs=True,
                      include_outputs=True,
                      include_data_types=None,
                      exclude_data_types=None,
                      include_values=True,
                      include_data_type_values=None,
                      exclude_data_type_values=None,
                      **kwargs):
    """
        Get all the resource scenarios for a given resource
        in a given scenario. If type_id is specified, only
        return the resource scenarios for the attributes
        within the type.
        args:
            ref_key (string): 'NETWORK', 'NODE', 'LINK', 'GROUP'
            ref_id (int): The ID of the network / node / link / group
            scenario_id (int): The ID of the scenario from which to get the resource scenarios
            type_id (int): A filter which limits the resource scenarios to just the attributes defined by the resource type
            expunge_session (bool): Expunge the DB session -- means that modifying the results will not update the database. Default True
            get_parent_data (bool): Return the data of the parent scenario of the requestsed scenario in addition to the specified scenario. Default False.
            include_inputs (bool) : Return resource scenarios which relate to resource attributes where the attr_is_var=N. Default True
            include_outputs (bool): Return resource scenarios which relate to resource attributes where the attr_is_var=Y. Default True
            include_data_types (list(string)): Return only resource scenarios with a dataset that has the type of one of these specified data types. Default None, meaning no filter is applied.
            exclude_data_types (list(string)): Return resource scenarios with a dataset that do NOT have the type of one of these specified data types. Default None, meaning no filter is applied.
            include_values (bool): Return the 'value' column of tDataset. Default True. Setting this to False can increase performance substantially due to the size of some dataset values.
            include_data_type_values (list(string)): When include_values is True, specify which dataset types should return with the value column included.
            exclude_data_type_values (list(string)): When include_values is True, specify which dataset types should return with the value column NOT included.

        returns:
            A list of JSONObjects representing Resource Scenarios, with a 'dataset' attribute and 'attribute' attribute.
    """

    user_id = kwargs.get('user_id')

    if ref_key is not None and ref_id is not None:
        resource_i = get_resource(ref_key, ref_id)
        resource_attributes = resource_i.attributes
    elif ref_key is None and ref_id is None:
        scenario = _get_scenario(scenario_id, user_id)
        resource_attributes = _get_all_network_resource_attributes(scenario.network_id)
    elif None in (ref_key, ref_id): # One of them is None
        raise HydraError("Unable to get data. Must specify a resource type (ref_key) and resource id (ref_id)")


    if include_inputs is False or include_outputs is False:
        if include_inputs is False:
            #only include outputs
            resource_attributes = list(filter(lambda x:x.attr_is_var=='Y', resource_attributes))

        if include_outputs is False:
            #only include inputs
            resource_attributes = list(filter(lambda x:x.attr_is_var=='N', resource_attributes))

    ra_ids = [ra.id for ra in resource_attributes]

    ra_map = dict((ra.id, ra) for ra in resource_attributes)

    scenario_i = _get_scenario(scenario_id, user_id)

    requested_rs = scenario_i.get_data(user_id,
                                       get_parent_data=get_parent_data,
                                       ra_ids=ra_ids,
                                       include_data_types=include_data_types,
                                       exclude_data_types=exclude_data_types,
                                       include_values=include_values,
                                       include_data_type_values=include_data_type_values,
                                       exclude_data_type_values=exclude_data_type_values)

    #map an raID to an rs for uses later
    ra_rs_map = {}
    for rs in requested_rs:
        ra_rs_map[rs.resource_attr_id] = rs

    #make a lookup table between an attr ID and an RS for filtering by types later.
    attr_rs_lookup = {} # Used later to remove results if they're not required
    for ra in resource_attributes:
        #Is there data for this RA?
        if ra_rs_map.get(ra.id) is not None:
            attr_rs_lookup[ra.attr_id] = ra_rs_map[ra.id]

    #Remove RS that are not defined by the specified type.
    if type_id is not None:

        type_limited_rs = []

        rs = db.DBSession.query(TypeAttr).filter(TypeAttr.type_id == type_id).all()
        for r in rs:
            type_limited_rs.append(attr_rs_lookup[r.attr_id])

        requested_rs = type_limited_rs

    if expunge_session is True:
        db.DBSession.expunge_all()

    return requested_rs

@required_perms("get_data", "get_network")
def get_attribute_datasets(attr_id, scenario_id, get_parent_data=False, **kwargs):
    """
        Retrieve all the datasets in a scenario for a given attribute.
        Also return the resource attributes so there is a reference to the node/link
    """

    user_id = kwargs.get('user_id')

    scenario_i = _get_scenario(scenario_id, user_id)

    try:
        a = db.DBSession.query(Attr).filter(Attr.id == attr_id).one()
    except NoResultFound:
        raise HydraError("Attribute %s not found"%(attr_id,))

    scenario_rs_i = scenario_i.get_data(user_id, get_parent_data=get_parent_data)

    #just in case the calling funciton hasn't cast this as as int
    attr_id = int(attr_id)

    requested_rs = []
    for rs_i in scenario_rs_i:
        if rs_i.resourceattr.attr_id == attr_id:
            #Finally add it to the list of RS to return
            requested_rs.append(rs_i)

    json_rs = []
    #Load the metadata too
    for rs in requested_rs:
        tmp_rs = JSONObject(rs)
        tmp_rs.resourceattr = JSONObject(rs.resourceattr)
        ra = tmp_rs.resourceattr
        if rs.resourceattr.node_id is not None:
            tmp_rs.resourceattr.node = get_resource(ra.ref_key, ra.node_id)
        elif rs.resourceattr.link_id is not None:
            tmp_rs.resourceattr.link = get_resource(ra.ref_key, ra.link_id)
        elif rs.resourceattr.group_id is not None:
            tmp_rs.resourceattr.resourcegroup = get_resource(ra.ref_key, ra.group_id)
        elif rs.resourceattr.network_id is not None:
            tmp_rs.resourceattr.network = get_resource(ra.ref_key, ra.network_id)

        json_rs.append(tmp_rs)


    return json_rs

@required_perms("get_data", "get_network")
def get_resourcegroupitems(group_id, scenario_id, get_parent_items=False, **kwargs):

    """
        Get all the items in a group, in a scenario. If group_id is None, return
        all items across all groups in the scenario.
    """

    user_id = kwargs.get('user_id')

    scenario_i = _get_scenario(scenario_id, user_id)

    requested_items = scenario_i.get_group_items(get_parent_items=get_parent_items)

    if group_id is not None:
        tmp_items = []
        for item_i in requested_items:
            if item_i.group_id == group_id:
                tmp_items.append(item_i)
        requested_items = tmp_items

    return requested_items

@required_perms("edit_network")
def delete_resourcegroupitems(scenario_id, item_ids, **kwargs):
    """
        Delete specified items in a group, in a scenario.
    """
    user_id = int(kwargs.get('user_id'))
    #check the scenario exists
    _get_scenario(scenario_id, user_id)
    for item_id in item_ids:
        rgi = db.DBSession.query(ResourceGroupItem).\
                filter(ResourceGroupItem.id == item_id).one()
        db.DBSession.delete(rgi)

    db.DBSession.flush()

@required_perms("edit_network")
def empty_group(group_id, scenario_id, **kwargs):
    """
        Delete all itemas in a group, in a scenario.
    """
    user_id = int(kwargs.get('user_id'))
    #check the scenario exists
    _get_scenario(scenario_id, user_id)

    rgi = db.DBSession.query(ResourceGroupItem).\
            filter(ResourceGroupItem.group_id == group_id).\
            filter(ResourceGroupItem.scenario_id == scenario_id).all()
    rgi.delete()

@required_perms("edit_network")
def add_resourcegroupitems(scenario_id, items, scenario=None, **kwargs):

    """
        Get all the items in a group, in a scenario.
    """
    user_id = int(kwargs.get('user_id'))

    if scenario is None:
        scenario = _get_scenario(scenario_id, user_id)

    _check_network_ownership(scenario.network_id, user_id)

    newitems = []
    for group_item in items:
        group_item_i = _add_resourcegroupitem(group_item, scenario.id)
        newitems.append(group_item_i)

    db.DBSession.flush()

    return newitems

def _add_resourcegroupitem(group_item, scenario_id):
    """
        Add a single resource group item (no DB flush, as it's an internal function)
    """
    if group_item.id and group_item.id > 0:
        try:
            group_item_i = db.DBSession.query(ResourceGroupItem).filter(ResourceGroupItem.id == group_item.id).one()
        except NoResultFound:
            raise ResourceNotFoundError("ResourceGroupItem %s not found" % (group_item.id))

    else:
        group_item_i = ResourceGroupItem()
        group_item_i.group_id = group_item.group_id
        if scenario_id is not None:
            group_item_i.scenario_id = scenario_id

        db.DBSession.add(group_item_i)

    ref_key = group_item.ref_key
    group_item_i.ref_key = ref_key
    if ref_key == 'NODE':
        group_item_i.node_id = group_item.ref_id if group_item.ref_id else group_item.node_id
    elif ref_key == 'LINK':
        group_item_i.link_id = group_item.ref_id if group_item.ref_id else group_item.link_id
    elif ref_key == 'GROUP':
        group_item_i.subgroup_id = group_item.ref_id if group_item.ref_id else group_item.subgroup_id

    return group_item_i

@required_perms("edit_data", "edit_network")
def update_value_from_mapping(source_resource_attr_id, target_resource_attr_id, source_scenario_id, target_scenario_id, **kwargs):
    """
        Using a resource attribute mapping, take the value from the source and apply
        it to the target. Both source and target scenarios must be specified (and therefor
        must exist).
    """
    user_id = int(kwargs.get('user_id'))

    rm = aliased(ResourceAttrMap, name='rm')
    #Check the mapping exists.
    mapping = db.DBSession.query(rm).filter(
        or_(
            and_(
                rm.resource_attr_id_a == source_resource_attr_id,
                rm.resource_attr_id_b == target_resource_attr_id
            ),
            and_(
                rm.resource_attr_id_a == target_resource_attr_id,
                rm.resource_attr_id_b == source_resource_attr_id
            )
        )
    ).first()

    if mapping is None:
        raise ResourceNotFoundError("Mapping between %s and %s not found"%
                                    (source_resource_attr_id,
                                     target_resource_attr_id))

    #check scenarios exist
    s1 = _get_scenario(source_scenario_id, user_id)
    s2 = _get_scenario(target_scenario_id, user_id)

    rs = aliased(ResourceScenario, name='rs')
    rs1 = db.DBSession.query(rs).filter(rs.resource_attr_id == source_resource_attr_id,
                                        rs.scenario_id == source_scenario_id).first()
    rs2 = db.DBSession.query(rs).filter(rs.resource_attr_id == target_resource_attr_id,
                                        rs.scenario_id == target_scenario_id).first()

    #3 possibilities worth considering:
    #1: Both RS exist, so update the target RS
    #2: Target RS does not exist, so create it with the dastaset from RS1
    #3: Source RS does not exist, so it must be removed from the target scenario if it exists
    return_value = None#Either return null or return a new or updated resource scenario
    if rs1 is not None:
        if rs2 is not None:
            log.info("Destination Resource Scenario exists. Updating dastaset ID")
            rs2.dataset_id = rs1.dataset_id
        else:
            log.info("Destination has no data, so making a new Resource Scenario")
            rs2 = ResourceScenario(
                resource_attr_id=target_resource_attr_id,
                scenario_id=target_scenario_id,
                dataset_id=rs1.dataset_id)
            db.DBSession.add(rs2)
        db.DBSession.flush()
        return_value = rs2
    else:
        log.info("Source Resource Scenario does not exist. Deleting destination Resource Scenario")
        if rs2 is not None:
            db.DBSession.delete(rs2)

    #lazy load the dataset
    return_value.dataset

    db.DBSession.flush()
    return return_value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
from ..base import *

from .resourceattr import ResourceAttr
from .resource import Resource

__all__ = ['ResourceGroup']

class ResourceGroup(Base, Inspect, Resource):
    """
    """

    __tablename__='tResourceGroup'
    __table_args__ = (
        UniqueConstraint('network_id', 'name', name="unique resourcegroup name"),
    )

    ref_key = 'GROUP'
    id = Column(Integer(), primary_key=True, nullable=False)
    name = Column(String(200),  nullable=False)
    description = Column(String(1000))
    status = Column(String(1),  nullable=False, server_default=text(u"'A'"))
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))
    network_id = Column(Integer(), ForeignKey('tNetwork.id'),  nullable=False)

    network = relationship('Network', backref=backref("resourcegroups", order_by=id, cascade="all, delete-orphan"), lazy='joined')

    _parents  = ['tNetwork']
    _children = ['tResourceAttr', 'tResourceType']

    def get_name(self):
        return self.group_name

    #For backward compatibility
    @property
    def group_id(self):
        return self.id

    @property
    def group_name(self):
        return self.name

    @group_name.setter
    def group_name_setter(self, value):
        self.name = value

    @property
    def group_description(self):
        return self.description

    @group_description.setter
    def group_description_setter(self):
        self.description = self.group_description

    def add_attribute(self, attr_id, attr_is_var='N'):
        res_attr = ResourceAttr()
        res_attr.attr_id = attr_id
        res_attr.attr_is_var = attr_is_var
        res_attr.ref_key = self.ref_key
        res_attr.group_id  = self.id
        self.attributes.append(res_attr)

        return res_attr

    def get_items(self, scenario_id):
        """
            Get all the items in this group, in the given scenario
        """
        items = get_session().query(ResourceGroupItem)\
                .filter(ResourceGroupItem.group_id==self.id).\
                filter(ResourceGroupItem.scenario_id==scenario_id).all()

        return items

    def check_read_permission(self, user_id, do_raise=True, is_admin=None):
        """
            Check whether this user can read this group
        """
        return self.network.check_read_permission(user_id, do_raise=do_raise, is_admin=is_admin)

    def check_write_permission(self, user_id, do_raise=True, is_admin=None):
        """
            Check whether this user can write this group
        """

        return self.network.check_write_permission(user_id, do_raise=do_raise, is_admin=is_admin)




import ctypes
import sys
import uuid


try:
    from ctypes import (
        windll,
        wintypes
    )
except ImportError as exc:
    raise Warning(f"Attempt to use Windows utils on {sys.platform} platform") from exc


# Pre-Vista CSIDL
CSIDL_COMMON_DOCUMENTS = 46
# Vista onwards Known Folder ID
# From https://learn.microsoft.com/en-us/windows/win32/shell/knownfolderid
KFP_PUBLIC_DOCUMENTS = uuid.UUID("{ED4824AF-DCE4-45A8-81E2-FC7965083634}")


class GUID(ctypes.Structure):
    # windsdk/guiddef.h
    _fields_ = [
        ("Data1", wintypes.DWORD),
        ("Data2", wintypes.WORD),
        ("Data3", wintypes.WORD),
        ("Data4", wintypes.BYTE*8)
    ]

    def __init__(self, _uuid):
        ctypes.Structure.__init__(self)
        self.Data1, self.Data2, self.Data3 = _uuid.fields[:3]
        self.Data4[:] = _uuid.bytes[8:]


def check_ret_err(result, func, args):
    if result != 0:
        raise OSerror("Unable to locate Common Documents")

    return args


def win_get_common_documents():
    # If the more recent KPF system is available, use this...
    if Win_SHGetKnownFolderPath := getattr(windll.shell32, "SHGetKnownFolderPath", None):
        Win_SHGetKnownFolderPath.argtypes = [
            ctypes.POINTER(GUID),
            wintypes.DWORD,
            wintypes.HANDLE,
            ctypes.POINTER(ctypes.c_wchar_p)
        ]
        Win_SHGetKnownFolderPath.errcheck = check_ret_err

        kfp = GUID(KFP_PUBLIC_DOCUMENTS)
        path_buf = ctypes.c_wchar_p()
        Win_SHGetKnownFolderPath(ctypes.byref(kfp), 0, 0, ctypes.byref(path_buf))

        return path_buf.value

    # ...otherwise use legacy CSIDL
    elif Win_SHGetFolderPathW := getattr(windll.shell32, "SHGetFolderPathW", None):
        Win_SHGetFolderPathW.argtypes = [
            wintypes.HWND,
            ctypes.c_int,
            wintypes.HANDLE,
            wintypes.DWORD,
            wintypes.LPCWSTR
        ]
        Win_SHGetFolderPathW.errcheck = check_ret_err

        path_buf = ctypes.create_unicode_buffer(wintypes.MAX_PATH)
        Win_SHGetFolderPathW(0, CSIDL_COMMON_DOCUMENTS, 0, 0, path_buf)

        return path_buf.value
    else:
        raise OSError("Unable to access Windows environment")


if __name__ == "__main__":
    cdp = win_get_common_documents()
    print(cdp)



# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2020 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import logging
import json
import pytest
from hydra_base.lib.objects import JSONObject
from hydra_base.exceptions import HydraError
log = logging.getLogger(__name__)


class TestTemplateInheritance:
    """
        TEMPLATES Functions
    """

    def test_add_child_template(self, client):
        """
            Test a child template of a template, then retrieve that
            and ensure that it is consistent with the parent.
        """

        template_j = client.testutils.create_template()

        child_template_j = client.testutils.create_child_template(template_j.id)

        child_template_received = client.get_template(child_template_j.id)

        assert template_j.id == child_template_j.parent_id == child_template_received.parent_id
        assert template_j.description == child_template_j.description
        assert child_template_j.description == child_template_received.description

    def test_add_type_to_child(self, client):
        """
            When adding a type to a child template, it should appear in the child
            but not the parent.
        """

        parent_template_j = client.testutils.create_template()

        child_template_j = client.testutils.create_child_template(parent_template_j.id)

        original_templatetypes = parent_template_j.templatetypes

        #Add a type to the child
        templatetype_j = client.testutils.create_templatetype(child_template_j.id)
        client.add_templatetype(templatetype_j)

        parent_template_received = client.get_template(parent_template_j.id)
        child_template_received = client.get_template(child_template_j.id)

        assert len(child_template_received.templatetypes)\
                == len(parent_template_received.templatetypes) + 1

        #Add a type to the parent
        templatetype_j = client.testutils.create_templatetype(parent_template_j.id)
        new_type_j = client.add_templatetype(templatetype_j)

        parent_template_received = client.get_template(parent_template_j.id)
        child_template_received = client.get_template(child_template_j.id)

        assert len(parent_template_received.templatetypes)\
                == len(original_templatetypes) + 1
        #check that one has been added to both parent and child
        #so we now have 2 more than originally
        assert len(child_template_received.templatetypes)\
                == len(original_templatetypes) + 2
        assert len(child_template_received.templatetypes)\
                == len(parent_template_received.templatetypes) + 1

    def test_update_type_in_child(self, client):
        """
            Test updating a template type in a child, which will result in creating
            a new template type object in the child template, which has a parent of the type
            in the parent template
        """

        parent_template_j = client.testutils.create_template()

        child_template_j = client.testutils.create_child_template(parent_template_j.id)

        child_template_received = client.get_template(child_template_j.id)

        #Store the types from the child for comparison later
        original_child_types = child_template_received.templatetypes


        #Get a type which inherits from the parent, and update it
        type_to_update = None
        for templatetype in child_template_received.templatetypes:
            if templatetype.template_id == parent_template_j.id:
                type_to_update = templatetype
                break


        child_type = client.add_child_templatetype(type_to_update.id, child_template_j.id)

        child_type.status = 'X'

        #hack to make sure the child attrs don't get added
        child_type['typeattrs'] = []

        client.update_templatetype(child_type)

        updated_child_template = client.get_template(child_template_j.id)

        #This should be one less than before
        assert len(updated_child_template.templatetypes) == len(original_child_types)

        for child_type in updated_child_template.templatetypes:
            if child_type.parent_id == type_to_update.id:
                assert child_type.status == 'X'


    def test_add_typeattr_to_child(self, client):
        """
            Test updating a template type in a child, by adding a type attribute
            to a type in the child template
        """

        parent_template_j = client.testutils.create_template()

        un_updated_parent_template = client.get_template(parent_template_j.id)

        child_template_j = client.testutils.create_child_template(parent_template_j.id)

        child_template_received = client.get_template(child_template_j.id)


        #Get a type which inherits from the parent, and update it
        type_to_update = None
        for templatetype in child_template_received.templatetypes:
            if templatetype.template_id == parent_template_j.id:
                type_to_update = templatetype
                break

        #Create a child of a parent *type*, so we can add a typeattr to it.
        #WHen the child template is requested, this type will contain all the
        #parent type attributes, as well as the newly added one to the child
        child_type = client.add_child_templatetype(type_to_update.id, child_template_j.id)

        #create some random attribute
        newattr = client.testutils.create_attribute('extra_attr', None)

        #Link this new attribute to the child type using a type attribute
        newtypeattr = JSONObject({
            'attr_id': newattr.id,
            'type_id': child_type.id
        })
        client.add_typeattr(newtypeattr)

        #Save the original types attributes for comparison later
        original_type_attrs = type_to_update.typeattrs

        #Fetch the child and parent templates to verify that the child has
        #been updated (with one extra type attr on one of its types) but the
        #parent has not been updated.
        updated_child_template = client.get_template(child_template_j.id)
        un_updated_parent_template = client.get_template(parent_template_j.id)

        updated_child_type_attrs = None
        updated_parent_type_attrs = None
        for templatetype in updated_child_template.templatetypes:
            if templatetype.parent_id == type_to_update.id:
                updated_child_type_attrs = templatetype.typeattrs
                break
        for templatetype in un_updated_parent_template.templatetypes:
            if templatetype.id == type_to_update.id:
                un_updated_parent_type_attrs = templatetype.typeattrs
                break

        assert len(updated_child_type_attrs) == len(original_type_attrs)+1
        assert len(updated_child_type_attrs) == len(un_updated_parent_type_attrs)+1

    def test_add_typeattr_to_child_then_parent(self, client):
        """
        Test for the situation where a typeattr is added to a child, then added to the parent.
        This is a valid case, and the behaviour should be that the parent typeattr should be ignored,
        rather than returning both or returning some merged typeattr.
        """

        parent_template_j = client.testutils.create_template()

        un_updated_parent_template = client.get_template(parent_template_j.id)

        child_template_j = client.testutils.create_child_template(parent_template_j.id)

        child_template_received = client.get_template(child_template_j.id)


        #Get a type which inherits from the parent, and update it
        type_to_update = None
        for templatetype in child_template_received.templatetypes:
            if templatetype.template_id == parent_template_j.id:
                type_to_update = templatetype
                break

        #Create a child of a parent *type*, so we can add a typeattr to it.
        #WHen the child template is requested, this type will contain all the
        #parent type attributes, as well as the newly added one to the child
        child_type = client.add_child_templatetype(type_to_update.id, child_template_j.id)

        #create some random attribute
        newattr = client.testutils.create_attribute('extra_attr', None)

        #Link this new attribute to the child type using a type attribute
        new_child_typeattr = JSONObject({
            'attr_id': newattr.id,
            'type_id': child_type.id
        })
        client.add_typeattr(new_child_typeattr)

        #Save the original types attributes for comparison later
        original_type_attrs = type_to_update.typeattrs

        #Fetch the child and parent templates to verify that the child has
        #been updated (with one extra type attr on one of its types) but the
        #parent has not been updated.
        updated_child_template = client.get_template(child_template_j.id)
        un_updated_parent_template = client.get_template(parent_template_j.id)

        updated_child_type_attrs = None
        updated_parent_type_attrs = None
        for templatetype in updated_child_template.templatetypes:
            if templatetype.parent_id == type_to_update.id:
                updated_child_type_attrs = templatetype.typeattrs
                break
        for templatetype in un_updated_parent_template.templatetypes:
            if templatetype.id == type_to_update.id:
                un_updated_parent_type_attrs = templatetype.typeattrs
                break

        assert len(updated_child_type_attrs) == len(original_type_attrs)+1
        assert len(updated_child_type_attrs) == len(un_updated_parent_type_attrs)+1


        #now add *the same* typeattr to the parent type. This should result in the child type returing its own typeattr
        #and the parent type returning a different typeattr, i.e. a different type_id with the same attr_id
        new_parent_typeattr = JSONObject({
            'attr_id': newattr.id,
            'type_id': type_to_update.id
        })
        client.add_typeattr(new_parent_typeattr)

        #Fetch both the parent and child to verify that the child and parent types
        #have typeattrs with the same attr_id but different type_ids
        updated_child_template = client.get_template(child_template_j.id)
        updated_parent_template = client.get_template(parent_template_j.id)

        updated_child_type_attrs = None
        updated_parent_type_attrs = None
        for templatetype in updated_child_template.templatetypes:
            if templatetype.parent_id == type_to_update.id:
                updated_child_type_attrs = templatetype.typeattrs
                break
        for templatetype in updated_parent_template.templatetypes:
            if templatetype.id == type_to_update.id:
                updated_parent_type_attrs = templatetype.typeattrs
                break

        #check the attr_ids are there
        assert newattr.id in [ta.attr_id for ta in updated_child_type_attrs]
        assert newattr.id in [ta.attr_id for ta in updated_parent_type_attrs]
        #check the type IDs are there
        assert child_type.id in [ta.type_id for ta in updated_child_type_attrs]
        assert type_to_update.id in [ta.type_id for ta in updated_parent_type_attrs]

        #check the type IDs are NOT in the other one's list of typeattrs. Specifically
        #we want to ensure that there are no duplicate typeattrs being returned
        assert child_type.id not in [ta.type_id for ta in updated_parent_type_attrs]
        assert type_to_update.id in [ta.type_id for ta in updated_child_type_attrs]

        #Make sure there are no duplicates returned explicitly
        assert len(set([ta.attr_id for ta in updated_child_type_attrs])) == len([ta.attr_id for ta in updated_child_type_attrs])
        assert len(set([ta.attr_id for ta in updated_parent_type_attrs])) == len([ta.attr_id for ta in updated_parent_type_attrs])




    def test_update_child_typeattr(self, client):
        """
            Test updating a template type attribute in a child, by adding a type attribute
            to a type in the child template, which has a parent in the parent template,
            and then setting the status of that to 'X'
        """

        parent_template_j = client.testutils.create_template()

        child_template_j = client.testutils.create_child_template(parent_template_j.id)

        child_template_received = client.get_template(child_template_j.id)

        #Get a type which inherits from the parent, and update it
        typeattr_to_update = None
        original_type = None
        for templatetype in child_template_received.templatetypes:
            if templatetype.template_id == parent_template_j.id:
                original_type = templatetype
                typeattr_to_update = templatetype.typeattrs[0]
                break

        #Create a child of a parent *typeattr*, so we can update it.
        #When the child template is requested, this typeattr will contain the combined
        #data from itself and from its parent.
        child_typeattr = client.add_child_typeattr(typeattr_to_update.id, child_template_j.id)

        child_typeattr.status = 'X'

        client.update_typeattr(child_typeattr)

        #Fetch the child and parent templates to verify that the child has
        #been updated (with one extra type attr on one of its types) but the
        #parent has not been updated.
        updated_child_template = client.get_template(child_template_j.id)
        un_updated_parent_template = client.get_template(parent_template_j.id)


        updated_child_type_attr = None
        update_child_type = None
        un_updated_parent_type_attr = None
        for templatetype in updated_child_template.templatetypes:
            for typeattr in templatetype.typeattrs:
                if typeattr.parent_id == typeattr_to_update.id:
                    updated_child_type = templatetype
                    updated_child_type_attr = typeattr
                    break
        for templatetype in un_updated_parent_template.templatetypes:
            for typeattr in templatetype.typeattrs:
                if typeattr.id == typeattr_to_update.id:
                    un_updated_parent_type_attr = typeattr
                    break

        #The parent and child type should have the same number of type attrs, but
        #one of the children's ones will be set to 'X'
        assert len(original_type.typeattrs) == len(updated_child_type.typeattrs)
        assert updated_child_type_attr.status == 'X'
        assert un_updated_parent_type_attr.status == 'A'

    def test_get_updated_type(self, client):
        """
            Test the individual 'get_type' functionality by
            updating a template type attribute in a child, by adding a type attribute
            to a type in the child template, which has a perent in the parent template,
            and then setting the status of that to 'X'
        """

        parent_template_j = client.testutils.create_template()

        child_template_j = client.testutils.create_child_template(parent_template_j.id)

        child_template_received = client.get_template(child_template_j.id)

        #Get a type which inherits from the parent, and update it
        typeattr_to_update = None
        original_type = None
        for templatetype in child_template_received.templatetypes:
            if templatetype.template_id == parent_template_j.id:
                original_type = templatetype
                typeattr_to_update = templatetype.typeattrs[0]
                break

        #Create a child of a parent *typeattr*, so we can update it.
        #When the child template is requested, this typeattr will contain the combined
        #data from itself and from its parent.
        child_typeattr = client.add_child_typeattr(typeattr_to_update.id, child_template_j.id)

        child_typeattr.status = 'X'

        client.update_typeattr(child_typeattr)

        #Fetch the child and parent templates to verify that the child has
        #been updated (with one extra type attr on one of its types) but the
        #parent has not been updated.
        updated_child_type = client.get_templatetype(child_typeattr.type_id)
        un_updated_parent_type = client.get_templatetype(typeattr_to_update.type_id)


        updated_child_type_attr = None
        update_child_type = None
        un_updated_parent_type_attr = None
        for typeattr in updated_child_type.typeattrs:
            if typeattr.parent_id == typeattr_to_update.id:
                updated_child_type = templatetype
                updated_child_type_attr = typeattr
                break
        for typeattr in un_updated_parent_type.typeattrs:
            if typeattr.id == typeattr_to_update.id:
                un_updated_parent_type_attr = typeattr
                break

        #The parent and child type should have the same number of type attrs, but
        #one of the children's ones will be set to 'X'
        assert len(original_type.typeattrs) == len(updated_child_type.typeattrs)
        assert updated_child_type_attr.status == 'X'
        assert un_updated_parent_type_attr.status == 'A'

    def test_get_updated_typeattr(self, client):
        """
            Test the individual 'get_type' functionality by
            updating a template type attribute in a child, by adding a type attribute
            to a type in the child template, which has a perent in the parent template,
            and then setting the status of that to 'X'
        """

        parent_template_j = client.testutils.create_template()

        child_template_j = client.testutils.create_child_template(parent_template_j.id)

        child_template_received = client.get_template(child_template_j.id)

        #Get a type which inherits from the parent, and update it
        typeattr_to_update = None
        original_type = None
        for templatetype in child_template_received.templatetypes:
            if templatetype.template_id == parent_template_j.id:
                original_type = templatetype
                typeattr_to_update = templatetype.typeattrs[0]
                break

        #Create a child of a parent *typeattr*, so we can update it.
        #When the child template is requested, this typeattr will contain the combined
        #data from itself and from its parent.
        child_typeattr = client.add_child_typeattr(typeattr_to_update.id, child_template_j.id)

        child_typeattr.status = 'X'

        client.update_typeattr(child_typeattr)

        updated_child_type_attr = client.get_typeattr(child_typeattr.id)
        un_updated_parent_type_attr = client.get_typeattr(typeattr_to_update.id)

        assert updated_child_type_attr.status == 'X'
        assert un_updated_parent_type_attr.status == 'A'


    def test_create_network_with_child_template(self, client):
        """
            This tests that a network is correctly linked with its template upon
            creation. The issues is:
            Upon creating using a template, a network is linked to a template with
            the template's 'network' template type. In an inherited template, this
            template type may belong to the parent, and so there is no direct means
            of linking a network to the correct child template, as the type ID of the 'network'
            template id will come from the parent. To mitigate this, a 'child template id' column
            exists on the tResourceType table which allows us to explicitly say which template was
            used to create the network.

            When creating a network using a child template, it's possible that
            there is no TemplateType entry for the network within the child template
            i.e. there's no entry in the DB for that template type.

            This means that when you request the template, the template type
            which is returned is that of the parent, and the type_id points to
            the template type entry in the parent template.

            If we store only this type_id, then we have lost the connection to
            the child template, as Hydra will only store the link to the type
            which is in the parent, and therefore hydra will think that the
            network was created using the parent template.

        """

        #first create a template
        parent_template_j = client.testutils.create_template()
        parent_network_type = list(filter(lambda x: x.resource_type=='NETWORK', parent_template_j.templatetypes))[0]

        #and a child
        child_template_j = client.testutils.create_child_template(parent_template_j.id)
        child_template_j = client.get_template(child_template_j.id)

        child_network_type = list(filter(lambda x: x.resource_type=='NETWORK', child_template_j.templatetypes))[0]

        #check that the child has a network type of its own
        assert child_network_type.id != parent_network_type.id
        assert child_network_type.parent_id == parent_network_type.id


        node_type = list(filter(lambda x: x.resource_type=='NODE', child_template_j.templatetypes))[0]
        #now create a network using the child's network type
        #The network type needs to be linked to the template ID of the
        #child so we can find the correct template

        project_j = client.add_project(JSONObject({'name': 'Template Inheritance Project'}))

        network = JSONObject({
            'project_id': project_j.id,
            'name': 'Test Network with Child Template',
            'types' : [{'id': child_network_type.id,
                        'child_template_id': child_template_j.id}],
            'nodes' : [{'name': 'Node1', 'x':0, 'y':0, 'types':[{'id':node_type.id}]}]
        })

        new_network = client.add_network(network)
        #Now do a get network to verify the network's there
        requested_network = client.get_network(new_network.id)

        assert requested_network.nodes[0].types[0].child_template_id == child_template_j.id

        #Make sure that the correct attributes have been added from the child (inherihted
        #from the parent)
        attr_ids_a = set(a.attr_id for a in requested_network.attributes)
        attr_ids_b = set(ta.attr_id for ta in parent_network_type.typeattrs)
        assert len(attr_ids_a.difference(attr_ids_b)) == 0


    def test_network_child_template_with_subclassed_type(self, client):
        """
        This tests a few operations involving making a derived type in a child template
        and verifies that the hydra_base.lib functions return a complete view of
        the resulting hierarchy.

          1. Create a Parent Template with 'Parent Node Type'
          2. Create a Child Template
          3. Create a Network using the Child Template
          4. Create a Child Template Node Type with parent type 'Parent Node Type'
          5. Create a Node instance of this type in the Network and verify correctness
          6. Create a Grandchild Template
          7. Create a Grandchild Template Node Type with parent type 'Child Template Node Type'
          8. Retrieve type information and verify the parent_ids attr contains both
             'Child Template Node Type' and 'Parent Node Type'
          9. Verify that the get_all_parent_types(ttype_id) lib function returns the
             correct Parent and GrandParent TemplateTypes

        """
        # Create Parent Template
        parent_template = client.testutils.create_template()

        # Create Child Template
        child_template = client.testutils.create_child_template(parent_template.id)
        child_template = client.get_template(child_template.id)

        child_network_type = list(filter(lambda x: x.resource_type=="NETWORK", child_template.templatetypes))[0]

        parent_node_type = list(filter(lambda x: x.resource_type=="NODE", child_template.templatetypes))[0]

        # Create Network using Child Template
        parent_node_name = "Original Parent Type Node"
        project_j = client.add_project(JSONObject({"name": "Child Template SubType Test Project"}))

        network = JSONObject({
            "project_id": project_j.id,
            "name": "Test Network with Child Template",
            "types" : [{"id": child_network_type.id,
                        "child_template_id": child_template.id}],
            "nodes" : [{"name": parent_node_name, "x": 0, "y": 0, "types":[{"id": parent_node_type.id}]}]
        })

        new_network = client.add_network(network)
        ret_network = client.get_network(new_network.id)

        # Add a subclassed Node type to the Child Template

        # 1. Make a Child Template Type
        templatetype = JSONObject()
        templatetype.name = "Child Template TT"
        templatetype.alias = "Child Template TT alias"
        templatetype.resource_type = "NODE"
        templatetype.template_id = child_template.id
        templatetype.parent_id = parent_node_type.id
        templatetype.layout = {"color": "green", "shapefile": "child_node.shp"}

        templatetype.typeattrs = []
        client.add_templatetype(templatetype)
        ret_tt = client.get_templatetype_by_name(child_template.id, "Child Template TT")

        # 2. Make a Node using the Child TemplateType
        node = JSONObject()
        node.id = -1
        node.name = "Child Template Added Node"
        node.description = "Child Template Node desc"
        node.x = 2
        node.y = 3

        node_types = []
        node_type = JSONObject()
        node_type.id  = ret_tt.id
        node_type.name = ret_tt.name

        node_types.append(node_type)
        node.types = node_types

        new_node = client.add_node(ret_network.id, node)

        # Verify Node has expected Type
        ret_child_node = client.get_node(new_node.id)
        ret_orig_node = client.get_node(ret_network.nodes[0].id)
        ret_child_template = client.get_template(child_template.id)
        ret_network = client.get_network(new_network.id)

        """
        ret_network now has Node with type.id==2, but
        ret_child_template.templatetypes has only Node type.id==6

        E.g. in the following, the original Parent Node Type (2) appears
        only in the parent_types, not in child_types even though these
        are the types of a Child Template.

        user_res_types = ("NODE", "LINK", "GROUP")
        ret_parent_template = client.get_template(parent_template.id)

        parent_types = set(tt.id for tt in ret_parent_template.templatetypes
                           if tt.resource_type.upper() in user_res_types)
        child_types = set(tt.id for tt in ret_child_template.templatetypes
                          if tt.resource_type.upper() in user_res_types)


        Now verify that the templates' parent_ids attr is complete...
        """

        # Make Grandchild Template with Node type derived from Child Node Type
        grandchild_template = client.testutils.create_child_template(ret_child_template.id)
        grandchild_template = client.get_template(grandchild_template.id)

        templatetype = JSONObject()
        templatetype.name = "Grandchild Template TT"
        templatetype.alias = "Grandchild Template TT alias"
        templatetype.resource_type = "NODE"
        templatetype.template_id = grandchild_template.id
        templatetype.parent_id = ret_child_node.types[0].type_id
        templatetype.layout = {"color": "blue", "shapefile": "grandchild_node.shp"}

        templatetype.typeattrs = []
        client.add_templatetype(templatetype)
        ret_tt = client.get_templatetype_by_name(grandchild_template.id, "Grandchild Template TT")

        # Retrieve the Grandchild template and verify both
        # 'Parent Node Type' and 'Child Node Type' are included
        # as parents of the 'Grandchild Node Type'
        ret_grandchild_template = client.get_template(grandchild_template.id)
        assert ret_orig_node.types[0].type_id in ret_grandchild_template.templatetypes[1].parent_ids
        assert ret_child_node.types[0].type_id in ret_grandchild_template.templatetypes[1].parent_ids
        # ...and no others...
        assert len(ret_grandchild_template.templatetypes[1].parent_ids) == 2

        # Verify that get_all_parent_types(ttype_id) returns the expected TemplateTypes...
        parents = client.get_all_parent_types(ret_grandchild_template.templatetypes[1].id)
        assert parents[0].id == ret_orig_node.types[0].type_id
        assert parents[1].id == ret_child_node.types[0].type_id
        # ...and no others...
        assert len(parents) == 2


    def test_delete_parent_type(self, client):
        """
            Test to ensure that when you delete a parent type, its child types
            are also deleted (or not depoending on the 'delete_children' flag)
        """
        #first create a template
        parent_template_j = client.testutils.create_template()
        parent_template_j = client.get_template(parent_template_j.id)
        parent_type = list(filter(lambda x: x.resource_type=='NODE', parent_template_j.templatetypes))[0]

        #and a child
        child_template_j = client.testutils.create_child_template(parent_template_j.id)
        child_template_j = client.get_template(child_template_j.id)

        child_type = client.add_child_templatetype(parent_type.id, child_template_j.id)


        #now delete the parent type, without force. This should error because there is a reference to it
        #from the child
        with pytest.raises(HydraError):
            client.delete_templatetype(parent_type.id)

        #now force the deletion
        client.delete_templatetype(parent_type.id, delete_children=True)


        updated_parent_template_j = client.get_template(parent_template_j.id)
        assert len(updated_parent_template_j.templatetypes) == len(parent_template_j.templatetypes) - 1
        assert child_type.id not in [t.id for t in updated_parent_template_j.templatetypes]


        updated_child_template_j = client.get_template(child_template_j.id)
        assert len(updated_child_template_j.templatetypes) == len(child_template_j.templatetypes) - 1
        assert child_type.id not in [t.id for t in updated_child_template_j.templatetypes]
        #all types in the child should now belong to the parent. The chid one is gone.
        for tt in updated_child_template_j.templatetypes:
            #there should always be a network type for the child
            if tt.resource_type == 'NETWORK':
                continue
            assert tt.resource_type is not None
            assert tt.name is not None
            assert tt.template_id==parent_template_j.id

    def test_get_child_template_by_name(self, client):

        #first create a template
        parent_template_j = client.testutils.create_template()
        parent_network_type = list(filter(lambda x: x.resource_type=='NETWORK', parent_template_j.templatetypes))[0]

        #and a child
        child_template_j = client.testutils.create_child_template(parent_template_j.id)
        child_template_j = client.get_template(child_template_j.id)


        child_template_2_j = client.get_template_by_name(child_template_j.name)

        assert len(parent_template_j.templatetypes) == len(child_template_j.templatetypes) == len(child_template_2_j.templatetypes)





#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
import logging
from sqlalchemy.orm.exc import NoResultFound
from hydra_base.lib.objects import JSONObject
from ..db.model import Rule, RuleTypeDefinition, RuleTypeLink, NetworkOwner, \
                       Network, Template, Project
from .. import db
from ..exceptions import HydraError, ResourceNotFoundError

from ..util.permissions import required_perms

LOG = logging.getLogger(__name__)

def _get_rule(rule_id, user_id, check_write=False):
    try:
        rule_i = db.DBSession.query(Rule).filter(Rule.id == rule_id).one()
    except NoResultFound:
        raise ResourceNotFoundError("Rule {0} not found".format(rule_id))

    #lazy load types
    rule_i.types

    rule_i.check_read_permission(user_id)

    if check_write is True:
        rule_i.check_write_permission(user_id)

    return rule_i

@required_perms("get_network", "get_rules")
def get_network_rules(network_id, summary=True, **kwargs):
    """
        Get all the rules within a network
        Args:
            network_id (int): ID of the resource
        Returns:
            List of Rule SQLAlchemy objects
    """
    user_id = kwargs.get('user_id')

    #just in case
    network_id = int(network_id)
    network = db.DBSession.query(Network).filter(Network.id==network_id).one()
    network.check_read_permission(user_id)

    rule_qry = db.DBSession.query(Rule).filter(Rule.status != 'X')

    #Go through the template hierarchy for all types defined on this network and extract
    #all rules associated to them.
    all_template_rules = []
    for rtype in network.types:
        if not hasattr(rtype, "template_id"):
            continue
        template = db.DBSession.query(Template).filter(Template.id==rtype.template_id).one()
        #need this to go top-bottom to apply rules from the top level down
        template_hierarchy = template.get_hierarchy().reverse()
        for current_template in template_hierarchy:
            this_template_rules = rule_qry.filter(Rule.template_id == current_template.id).all()
            all_template_rules = all_template_rules + this_template_rules


    network_rules = rule_qry.filter(Rule.network_id == network_id).all()
    all_network_rules = network_rules + all_template_rules

    #lazy load types
    if summary is False:
        for rule in all_network_rules:
            rule.types

    return all_network_rules


@required_perms("get_template", "get_rules")
def get_template_rules(template_id, summary=True, **kwargs):
    """
       Retrieve all Rules defined in a template
       args:
         <template_id>: integer (or castable) identifying a template
         <summary>: bool specifying whether type-loading should be
                    omitted
       return:
         list of Rule SQLAlchemy objects
    """
    template_id = int(template_id)
    template = db.DBSession.query(Template).filter(Template.id==template_id).one()
    rule_qry = db.DBSession.query(Rule).filter(Rule.status != 'X')

    all_template_rules = rule_qry.filter(Rule.template_id == template_id).all()
    template_hierarchy = template.get_hierarchy(**kwargs).reverse()
    if template_hierarchy is not None:
        for current_template in template_hierarchy:
            this_template_rules = rule_qry.filter(Rule.template_id == current_template.id).all()
            all_template_rules = all_template_rules + this_template_rules

    if summary is False:
        for rule in all_template_rules:
            rule.types

    return all_template_rules


@required_perms("get_project", "get_rules")
def get_project_rules(project_id, summary=True, **kwargs):
    """
       Retrieve all Rules defined in a project
       args:
         <project_id>: integer (or castable) identifying a project
         <summary>: bool specifying whether type-loading should be
                    omitted
       return:
         list of Rule SQLAlchemy objects
    """
    user_id = kwargs.get("user_id")

    project_id = int(project_id)
    project = db.DBSession.query(Project).filter(Project.id==project_id).one()
    project.check_read_permission(user_id)

    rule_qry = db.DBSession.query(Rule).filter(Rule.status != 'X')

    project_rules = rule_qry.filter(Rule.project_id == project.id).all()

    if summary is False:
        for rule in project_rules:
            rule.types

    return project_rules


@required_perms("get_rules")
def get_resource_rules(ref_key, ref_id, summary=True, **kwargs):
    """
        Get all the rules for a given resource.
        Args:
            ref_key (string): NETWORK, PROJECT, TEMPLATE
            ref_id (int): ID of the resource
        Returns:
            List of Rule SQLAlchemy objects
    """
    ref_key == ref_key.upper()

    if ref_key.upper() == 'NETWORK':
        ret_func = get_network_rules
    elif ref_key.upper() == 'PROJECT':
        ret_func = get_project_rules
    elif ref_key.upper() == 'TEMPLATE':
        ret_func = get_template_rules
    else:
        raise HydraError("Ref Key {0} not recognised.".format(ref_key))

    rules = ret_func(ref_id, summary=summary, **kwargs)

    if summary is False:
        #lazy load types
        for rule in rules:
            rule.types

    return rules


@required_perms("get_rules")
def get_rules_of_type(typecode, scenario_id=None, **kwargs):
    """
        Get all the rules for a given resource.
        Args:
            ref_key (string): NETWORK, PROJECT, TEMPLATE
            ref_id (int): ID of the resource
            scenario_id (int): Optional which filters on scenario ID also
        Returns:
            List of Rule SQLAlchemy objects
    """

    user_id = kwargs.get('user_id')

    rule_qry = db.DBSession.query(Rule).filter(Rule.status == 'A')\
                                        .join(RuleTypeLink)\
                                        .filter(RuleTypeLink.code == typecode)

    rule_qry = rule_qry.join(NetworkOwner, Rule.network_id==NetworkOwner.network_id).filter(NetworkOwner.user_id == int(user_id))

    if scenario_id is not None:
        rule_qry = rule_qry.filter(Rule.scenario_id == scenario_id)

    rules = rule_qry.all()

    return rules

@required_perms("get_rules")
def get_rule(rule_id, **kwargs):
    """
        Get a rule by its ID
        Args:
            rule_id (int); The ID of the rule
        Returns:
            A SQLAlchemy ORM Rule object
    """
    user_id = kwargs.get('user_id')
    rule = _get_rule(rule_id, user_id)
    return rule

@required_perms("add_rules")
def add_rules(rules, **kwargs):
    LOG.info("Adding %s rules."%len(rules))
    for rule in rules:
        add_rule(rule, **kwargs)

    LOG.info("Rules Added")

@required_perms("add_rules")
def add_rule(rule, **kwargs):
    """
        Add a new rule.
        Args:
            rule: A JSONObject or Spyne Rule object
        Returns
            A SQLAlchemy ORM Rule object
    """

    user_id = kwargs.get('user_id')

    rule_i = Rule()
    rule_i.ref_key = rule.ref_key
    if rule.ref_key.upper() == 'NETWORK':
        rule_i.network_id = rule.network_id if rule.network_id else rule.ref_id
    elif rule.ref_key.upper() == 'PROJECT':
        rule_i.project_id = rule.project_id if rule.project_id else rule.ref_id
    elif rule.ref_key.upper() == 'TEMPLATE':
        rule_i.template_id = rule.template_id if rule.template_id else rule.ref_id
    else:
        raise HydraError("Ref Key {0} not recognised.".format(rule.ref_key))

    rule_i.name = rule.name
    rule_i.description = rule.description
    rule_i.value = rule.value
    rule_i.format = rule.format

    rule_i.set_types(rule.types)

    db.DBSession.add(rule_i)

    db.DBSession.flush()

    return rule_i

@required_perms("update_rules")
def update_rule(rule, **kwargs):
    """
        Add a new rule.
        Args:
            rule: A JSONObject or Spyne Rule object
        Returns
            A SQLAlchemy ORM Rule object
    """

    user_id = kwargs.get('user_id')

    rule_i = _get_rule(rule.id, user_id, check_write=True)

    rule_i.ref_key = rule.ref_key
    if rule.ref_key.upper() == 'NETWORK':
        rule_i.network_id = rule.network_id if rule.network_id else rule.ref_id
    elif rule.ref_key.upper() == 'PROJECT':
        rule_i.network_id = rule.project_id if rule.project_id else rule.ref_id
    elif rule.ref_key.upper() == 'TEMPLATE':
        rule_i.network_id = rule.template_id if rule.template_id else rule.ref_id
    else:
        raise HydraError("Ref Key {0} not recognised.".format(rule.ref_key))

    rule_i.name = rule.name
    rule_i.description = rule.description
    rule_i.format = rule.format
    rule_i.value = rule.value
    rule_i.status = rule.status

    rule_i.set_types(rule.types)

    db.DBSession.flush()

    return rule_i

@required_perms("update_rules")
def set_rule_type(rule_id, typecode, **kwargs):
    """
        Assign a rule type to a rule
        args:
            rule_id (int): THe ID of the rule to apply the type to
            typecode (string): Types do not use IDS as identifiers, rather codes.
                               Apply the type with this code to the rule
        returns:
            The updated rule (Sqlalchemy ORM Object)
    """
    user_id = kwargs.get('user_id')

    rule_i = _get_rule(rule_id, user_id, check_write=True)

    ruletypelink = RuleTypeLink()
    ruletypelink.code = typecode

    rule_i.types.append(ruletypelink)

    db.DBSession.flush()

    return rule_i

@required_perms("edit_network", "get_rules", "add_rules")
def clone_resource_rules(ref_key, ref_id, target_ref_key=None, target_ref_id=None, **kwargs):
    """
        Clone a rule
        args:
            ref_key (int): NETWORK, PROJECT, TEMPLATE
            ref_id (int): The ID of the relevant resource
            target_ref_key (string): If the rule is to be cloned into a
                                     different resource, specify the new resources type
            target_ref_id (int): If the rule is to be cloned into a different
                                 resources, specify the resource ID.

        Cloning will only occur into a different resource if both
        ref_key AND ref_id are provided. Otherwise it will maintain its
        original ref_key and ref_id.

        returns:
            list of rule ORM objects.
    """

    user_id = kwargs.get('user_id')

    resource_rules = get_resource_rules(ref_key, ref_id, user_id=user_id)

    cloned_rules = []

    for rule in resource_rules:
        cloned_rules.append(clone_rule(rule.id,
                                       target_ref_key=target_ref_key,
                                       target_ref_id=target_ref_id,
                                       user_id=user_id))

    return cloned_rules

@required_perms("edit_network", "get_rules", "add_rules")
def clone_rule(rule_id, target_ref_key=None, target_ref_id=None, **kwargs):
    """
        Clone a rule
        args:
            rule_id (int): The rule to clone
            target_ref_key (string): If the rule is to be cloned into a different
                                     resource, specify the new resources type
            target_ref_id (int): If the rule is to be cloned into a
                                 different resources, specify the resource ID.
        Cloning will only occur into a different resource if both ref_key AND
        ref_id are provided. Otherwise it will maintain its original ref_key and ref_id.

        return:
            SQLAlchemy ORM object
    """


    user_id = kwargs.get('user_id')

    rule_i = _get_rule(rule_id, user_id, check_write=True)

    #lazy load types
    rule_i.types

    rule_j = JSONObject(rule_i)
    rule_j.owners = rule_i.get_owners()

    #Unset the reference ID for the rule in case the target resource type
    #has changed, then apply the new ref_key and ref_id
    if target_ref_key is not None and target_ref_id is not None:
        rule_j.network_id = None
        rule_j.project_id = None
        rule_j.template_id = None
        rule_j.ref_key = target_ref_key
        if target_ref_key == 'NETWORK':
            rule_j.network_id = target_ref_id
        elif target_ref_key == 'PROJECT':
            rule_j.project_id = target_ref_id
        elif target_ref_key == 'TEMPLATE':
            rule_j.template_id = target_ref_id

    cloned_rule = add_rule(rule_j, **kwargs)

    return cloned_rule

@required_perms("update_rules")
def delete_rule(rule_id, **kwargs):
    """
        Set the status of a rule to 'X'
        args:
            rule_id: The id to update
        returns:
            None
    """

    user_id = kwargs.get('user_id')
    rule_i = _get_rule(rule_id, user_id, check_write=True)

    rule_i.status = 'X'

    db.DBSession.flush()

@required_perms("update_rules")
def activate_rule(rule_id, **kwargs):
    """
        Set the status of a rule to 'A'
        args:
            rule_id: The id to update
        returns:
            None
    """

    user_id = kwargs.get('user_id')
    rule_i = _get_rule(rule_id, user_id, check_write=True)

    rule_i.status = 'A'

    db.DBSession.flush()

@required_perms("delete_rules")
def purge_rule(rule_id, **kwargs):
    """
        Remove a rule from the DB permanently
        args:
            rule_id: The id to purge
        returns:
            None
    """
    user_id = kwargs.get('user_id')

    rule_i = _get_rule(rule_id, user_id, check_write=True)

    db.DBSession.delete(rule_i)
    db.DBSession.flush()

@required_perms("update_rules")
def add_rule_type_definition(ruletypedefinition, **kwargs):
    """
        Add a rule type definition
        Args:
            ruletypedefinition (Spyne or JSONObject object). This looks like:
                                {
                                  'name': 'My Rule Type',
                                  'code': 'my_rule_type'
                                }
        Returns:
            ruletype_i (SQLAlchemy ORM Object) new rule type from DB
    """
    rule_type_i = RuleTypeDefinition()

    rule_type_i.code = ruletypedefinition.code
    rule_type_i.name = ruletypedefinition.name

    existing_rtd = db.DBSession.query(RuleTypeDefinition).filter(
        RuleTypeDefinition.code == ruletypedefinition.code).first()

    if existing_rtd is None:
        db.DBSession.add(rule_type_i)

        db.DBSession.flush()

        return rule_type_i
    else:
        return existing_rtd

@required_perms("get_rules")
def get_rule_type_definitions(**kwargs):
    """
        Get all rule types
        args: None
        returns:
            List of SQLAlchemy objects.
    """

    all_rule_type_definitions_i = db.DBSession.query(RuleTypeDefinition).all()

    return all_rule_type_definitions_i

@required_perms("get_rules")
def get_rule_type_definition(typecode, **kwargs):
    """
        Get a Type with the given typecode
        args:
            typecode: Type definitions do not used IDs, rather codes. This is to code to retrieve
        returns:
            SQLAlchemy ORM Object
        raises:
            ResourceNotFoundError if the rule type definintion code does not exist
    """
    try:
        rule_type_definition_i = db.DBSession.query(RuleTypeDefinition)\
            .filter(RuleTypeDefinition.code == typecode).one()
    except NoResultFound:
        raise ResourceNotFoundError("Rule Type Definition {0} not found".format(typecode))

    return rule_type_definition_i

@required_perms("update_rules")
def purge_rule_type_definition(typecode, **kwargs):
    """
        Delete a rule type from the DB. Doing this will revert all existing rules to
        having no type (rather than deleting them)
        args:
            typecode: Type definitions do not used IDs, rather codes. This is to code to purge
        returns:
             None
        raises:
            ResourceNotFoundError if the rule type definintion code does not exist
    """
    try:
        rule_type_definition_i = db.DBSession.query(RuleTypeDefinition)\
            .filter(RuleTypeDefinition.code == typecode).one()
    except NoResultFound:
        raise ResourceNotFoundError("Rule Type {0} not found.".format(typecode))

    db.DBSession.delete(rule_type_definition_i)

    db.DBSession.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

from ..exceptions import HydraError, ResourceNotFoundError
from . import scenario, network

from .. import db
from ..db.model import ResourceGroup, ResourceGroupItem, Node, Link
from .scenario import _get_scenario
from sqlalchemy.orm.exc import NoResultFound

import logging
log = logging.getLogger(__name__)

def _get_group(group_id):
    try:
        return db.DBSession.query(ResourceGroup).filter(ResourceGroup.id==group_id).one()
    except NoResultFound:
        raise ResourceNotFoundError("ResourceGroup %s not found"%(group_id,))

def _get_item(item_id):
    try:
        item = db.DBSession.query(ResourceGroupItem).filter(ResourceGroupItem.id==item_id).one()
        return item
    except NoResultFound:
        raise ResourceNotFoundError("ResourceGroupItem %s not found"%(item_id,))

def add_resourcegroup(group, network_id,**kwargs):
    """
        Add a new group to a network.
    """
    group_i                   = ResourceGroup()
    group_i.name              = group.name
    group_i.description       = group.description
    group_i.status            = group.status
    group_i.network_id        = network_id
    db.DBSession.add(group_i)
    db.DBSession.flush()
    return group_i

def delete_resourcegroup(group_id, purge_data='N', **kwargs):
    """
        Add a new group to a scenario.
    """
    group_i = _get_group(group_id)

    if purge_data == 'Y':
        network._purge_datasets_unique_to_resource('GROUP', group_id)
    #This should cascaded to delete all the group items.
    db.DBSession.delete(group_i)
    db.DBSession.flush()

    return 'OK'

def update_resourcegroup(group,**kwargs):
    """
        Add a new group to a network.
    """

    group_i                 = _get_group(group.id)
    group_i.name            = group.name
    group_i.description     = group.description
    group_i.status          = group.status

    db.DBSession.flush()

    return group_i


def add_resourcegroupitem(group_item, scenario_id,**kwargs):

    _get_scenario(scenario_id, kwargs['user_id'], check_can_edit=True)
    #Check whether the ref_id is correct.

    if group_item.ref_key == 'NODE':
        try:
            db.DBSession.query(Node).filter(Node.id==group_item.ref_id).one()

        except NoResultFound:
            raise HydraError("Invalid ref ID %s for a Node group item!"%(group_item.ref_id))
    elif group_item.ref_key == 'LINK':
        try:
            db.DBSession.query(Link).filter(Link.id==group_item.ref_id).one()
        except NoResultFound:
            raise HydraError("Invalid ref ID %s for a Link group item!"%(group_item.ref_id))
    elif group_item.ref_key == 'GROUP':
        try:
            db.DBSession.query(ResourceGroup).filter(ResourceGroup.id==group_item.ref_id).one()
        except NoResultFound:
            raise HydraError("Invalid ref ID %s for a Group group item!"%(group_item.ref_id))
    else:
        raise HydraError("Invalid ref key: %s"%(group_item.ref_key))

    group_item_i             = ResourceGroupItem()
    group_item_i.scenario_id = scenario_id
    group_item_i.group_id    = group_item.group_id
    group_item_i.ref_key     = group_item.ref_key
    if group_item.ref_key == 'NODE':
        group_item_i.node_id      = group_item.ref_id
    elif group_item.ref_key == 'LINK':
        group_item_i.link_id = group_item.ref_id
    elif group_item.ref_key == 'GROUP':
        group_item_i.subgroup_id = group_item.ref_id


    db.DBSession.add(group_item_i)
    db.DBSession.flush()

    return group_item_i

def delete_resourcegroupitem(item_id,**kwargs):
    group_item_i = _get_item(item_id)

    _get_scenario(group_item_i.scenario_id, kwargs['user_id'], check_can_edit=True)
    db.DBSession.delete(group_item_i)
    db.DBSession.flush()

    return 'OK'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

class HydraError(Exception):
    __error_code__ = None

    def __init__(self, message="A hydra error has occurred"):
        # Call the base class constructor with the parameters it needs
        self.code = error_codes.get(self.__error_code__, '0000')
        self.message = message # Added for Py3
        super(HydraError, self).__init__({"message": message, "code": self.code})



class HydraDBError(HydraError):
    __error_code__ = 'HYDRADB'


class HydraPluginError(HydraError):
    __error_code__ = 'HYDRAPLUGIN'


class ResourceNotFoundError(HydraError):
    __error_code__ = 'HYDRARESOURCE'


class HydraAttributeError(HydraError):
    __error_code__ = 'HYDRAATTR'


class PermissionError(HydraError):
    __error_code__ = 'HYDRAPERM'


class OwnershipError(HydraError):
    __error_code__ = 'HYDRAOWNER'


class DataError(HydraError):
    __error_code__ = 'HYDRADATA'


class ValidationError(HydraError):
    pass


class HydraLoginUserNotFound(HydraError):
    __error_code__ = 'HYDRA_LOGIN_USER_NOT_FOUND'

class HydraLoginUserMaxAttemptsExceeded(HydraError):
    __error_code__ = 'HYDRA_LOGIN_USER_MAX_ATTEMPTS_EXCEEDED'

class HydraLoginUserPasswordWrong(HydraError):
    __error_code__ = 'HYDRA_LOGIN_USER_PASSWORD_WRONG'

#
#ERROR CODES FOR HYDRA
#Categories are:
#DB Errors:         100 - 199
#Plugin Errors:     200 - 299
#ResourceErrors:    300 - 399
#Attribute Errors:  400 - 499
#Permission Errors: 500 - 599
#Data Errors        600 - 699
#Ownership Errors   700 - 799
#Login Errors       800 - 899
#
error_codes = {
    'HYDRADB'               : "100",
    'HYDRAPLUGIN'           : "200",
    'HYDRARESOURCE'         : "300",
    'HYDRAATTR'             : "400",
    'HYDRAPERM'             : "500",
    'HYDRADATA'             : "600",
    'HYDRAOWNER'            : "700",
    'HYDRA_LOGIN_USER_NOT_FOUND'  : "800",
    'HYDRA_LOGIN_USER_MAX_ATTEMPTS_EXCEEDED'  : "801",
    'HYDRA_LOGIN_USER_PASSWORD_WRONG'  : "802"
}
